# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
import copy
from typing import Optional, Tuple
import numpy as np
import os
from scipy.optimize import minimize
from scipy.optimize import LinearConstraint

from grid2op.dtypes import dt_float
from grid2op.Environment import BaseEnv
from grid2op.Action import BaseAction
from grid2op.Backend import Backend
from grid2op.Observation.baseObservation import BaseObservation
from grid2op.Observation.highresSimCounter import HighResSimCounter
from grid2op.Exceptions import SimulatorError, InvalidRedispatching


class Simulator(object):
    """This class represents a "simulator". It allows to check the impact on this or that on th powergrid, quite 
    like what human operators have at their disposal in control rooms.
    
    It behaves similarly to `env.step(...)` or `obs.simulate(...)` with a few key differences:
    
    - you can "chain" the call to simulator: `simulator.predict(...).predict(...).predict(...)`
    - it does not take into account the "time": no cooldown on lines nor substation, storage
      "state of charge" (energy) does not decrease when you use them
    - no automatic line disconnection: lines are not disconnected when they are above their limit
    - no opponent will act on the grid
    
    Please see the documentation for usage examples.
    
    """
    def __init__(
        self, backend: Optional[Backend], env: Optional[BaseEnv] = None, tol_redisp=1e-6,
        _highres_sim_counter: Optional[HighResSimCounter] =None
    ):
        # backend should be initiliazed !
        if backend is not None:
            if not isinstance(backend, Backend):
                raise SimulatorError(
                    f'The "backend" argument should be an object '
                    f'of type "Backend" you provided {backend}'
                )
            if env is not None:
                raise SimulatorError(
                    "When building a simulator with a grid2op backend "
                    'make sure you set the kwarg "env=None"'
                )
            if backend._can_be_copied:
                self.backend: Backend = backend.copy()
            else:
                raise SimulatorError("Impossible to make a Simulator when you "
                                     "cannot copy the backend.")
        else:
            if env is None:
                raise SimulatorError(
                    "If you want to build a simulator with a blank / None "
                    'backend you should provide an environment (kwargs "env")'
                )
            if not isinstance(env, BaseEnv):
                raise SimulatorError(
                    f"Make sure the environment you provided is "
                    f"a grid2op Environment (an object of a type "
                    f"inheriting from BaseEnv"
                )
            if env.backend._can_be_copied:
                self.backend = env.backend.copy()
            else:
                raise SimulatorError("Impossible to make a Simulator when you "
                                     "cannot copy the backend of the environment.")

        self.current_obs: BaseObservation = None
        self._converged: Optional[bool] = None
        self._error: Optional[Exception] = None

        self._tol_redisp: float = tol_redisp
        if _highres_sim_counter is not None:
            self._highres_sim_counter = _highres_sim_counter
        else:
            self._highres_sim_counter = HighResSimCounter()

    @property
    def converged(self) -> bool:
        """
        
        Returns
        -------
        bool
            Whether or not the powerflow has converged
        """
        return self._converged

    @converged.setter
    def converged(self, values):
        raise SimulatorError("Cannot set this property.")

    def copy(self) -> "Simulator":
        """Allows to perform a (deep) copy of the simulator.

        Returns
        -------
        Simulator
            A (deep) copy of the simulator you want to copy.

        Raises
        ------
        SimulatorError
            In case the simulator is not initialized.
            
        """
        if self.current_obs is None:
            raise SimulatorError(
                "Impossible to copy a non initialized Simulator. "
                "Have you used `simulator.set_state(obs, ...)` with a valid observation before ?"
            )
        res = copy.copy(self)
        res.backend = res.backend.copy()
        res.current_obs = res.current_obs.copy()
        # do not copy this !
        res._highres_sim_counter = self._highres_sim_counter
        return res

    def change_backend(self, backend: Backend):
        """You can use this function in case you want to change the "solver" use to perform the computation.
        
        For example, you could use a machine learning based model to do the computation (to accelerate them), provided
        that you have at your disposal such an algorithm. 

        .. warning::
            The backend you pass as argument should be initialized with the same grid as the one currently in use.
        
        Notes
        -----
        Once changed, all the "simulator" that "derived" from this simulator will use the same backend types.
        
        Parameters
        ----------
        backend : Backend
            Another grid2op backend you can use to perform the computation.

        Raises
        ------
        SimulatorError
            When you do not pass a correct backend.
            
        """
        if not isinstance(backend, Backend):
            raise SimulatorError(
                "when using change_backend function, the backend should"
                " be an object (an not a class) of type backend"
            )
        self.backend.close()
        self.backend = backend.copy()  # backend_class.init_grid(type(self.backend))
        self.set_state(obs=self.current_obs)

    def change_backend_type(self, backend_type: type, grid_path: os.PathLike, **kwargs):
        """It allows to change the type of the backend used

        Parameters
        ----------
        backend_type : type
            The new backend type
        grid_path : os.PathLike
            The path from where to load the powergrid
        kwargs:
            Extra arguments used to build the backend.
            
        Notes
        -----
        Once changed, all the "simulator" that "derived" from this simulator will use the same backend types.
        
        Raises
        ------
        SimulatorError
            if something went wrong (eg you do not pass a type, your type does not inherit from Backend, the file
            located at `grid_path` does not exists etc.)
        """
        if not isinstance(backend_type, type):
            raise SimulatorError(
                "when using change_backend_type function, the backend_type should"
                " be a class an not an object"
            )
        if not issubclass(backend_type, Backend):
            raise SimulatorError(
                "when using change_backend_type function, the backend_type should"
                " be subtype of class Backend"
            )
        if not os.path.exists(grid_path):
            raise SimulatorError(
                f'the supposed grid path "{grid_path}" does not exists'
            )
        if not os.path.isfile(grid_path):
            raise SimulatorError(f'the supposed grid path "{grid_path}" if not a file')

        tmp_backend = backend_type(**kwargs)
        tmp_backend.load_grid(grid_path)
        tmp_backend.assert_grid_correct()
        self.backend.close()
        self.backend = tmp_backend
        self.set_state(obs=self.current_obs)

    def set_state(
        self,
        obs: Optional[BaseObservation] = None,
        do_powerflow: bool = True,
        new_gen_p: np.ndarray = None,
        new_gen_v: np.ndarray = None,
        new_load_p: np.ndarray = None,
        new_load_q: np.ndarray = None,
        update_thermal_limit: bool = True,
    ):
        """Set the state of the simulator to a given state described by an observation (and optionally some
        new loads and generation)

        Parameters
        ----------
        obs : Optional[BaseObservation], optional
            The observation to get the state from, by default None
        do_powerflow : bool, optional
            Whether to use the underlying backend to get a consistent state after
            this modification or not, by default True
        new_gen_p : np.ndarray, optional
            new generator active setpoint, by default None
        new_gen_v : np.ndarray, optional
            new generator voltage setpoint, by default None
        new_load_p : np.ndarray, optional
            new load active consumption, by default None
        new_load_q : np.ndarray, optional
            new load reactive consumption, by default None
        update_thermal_limit: bool, optional
            Do you update the thermal limit of the backend (we recommend to leave it to `True`
            otherwise some bugs can appear such as 
            https://github.com/rte-france/Grid2Op/issues/377)

        Raises
        ------
        SimulatorError
            In case the current simulator is not initialized.
        """

        if obs is not None:
            self.current_obs = obs.copy()

        if self.current_obs is None:
            raise SimulatorError(
                "The simulator is not initialized. Have you used `simulator.set_state(obs, ...)` with a valid observation before ?"
            )

        # you cannot use "simulate" of the observation in this class
        self.current_obs._obs_env = None
        self.current_obs._forecasted_inj = []
        self.current_obs._forecasted_grid = []

        # udpate the new state if needed
        if new_load_p is not None:
            self.current_obs.load_p[:] = new_load_p
        if new_load_q is not None:
            self.current_obs.load_q[:] = new_load_q

        if new_gen_p is not None:
            self.current_obs.gen_p[:] = new_gen_p
        if new_gen_v is not None:
            self.current_obs.gen_v[:] = new_gen_v

        self._converged = None
        self.error = None
        self.backend.update_from_obs(self.current_obs, force_update=True)

        if update_thermal_limit:
            self.backend.update_thermal_limit_from_vect(self.current_obs.thermal_limit)
            
        if do_powerflow:
            self._do_powerflow()

    def _do_powerflow(self):
        self._highres_sim_counter.add_one()
        self._converged, self._error = self.backend.runpf()

    def _update_obs(self):
        if self._converged:
            self.current_obs._update_attr_backend(self.backend)
        else:
            self.current_obs.set_game_over()

    def _adjust_controlable_gen(
        self, new_gen_p: np.ndarray, target_dispatch: np.ndarray, sum_target: float
    ) -> Optional[float]:
        nb_dispatchable = self.current_obs.gen_redispatchable.sum()

        # which generators needs to be "optimized" -> the one where
        # the target function matter
        gen_in_target = target_dispatch[self.current_obs.gen_redispatchable] != 0.0

        # compute the upper / lower bounds for the generators
        dispatchable = new_gen_p[self.current_obs.gen_redispatchable]
        val_min = (
            self.current_obs.gen_pmin[self.current_obs.gen_redispatchable]
            - dispatchable
        )
        val_max = (
            self.current_obs.gen_pmax[self.current_obs.gen_redispatchable]
            - dispatchable
        )

        # define the target function (things that will be minimized)
        target_dispatch_redisp = target_dispatch[self.current_obs.gen_redispatchable]

        coeffs = 1.0 / (
            self.current_obs.gen_max_ramp_up
            + self.current_obs.gen_max_ramp_down
            + self._tol_redisp
        )
        weights = np.ones(nb_dispatchable) * coeffs[self.current_obs.gen_redispatchable]
        weights /= weights.sum()

        scale_objective = max(0.5 * np.abs(target_dispatch_redisp).sum() ** 2, 1.0)
        scale_objective = np.round(scale_objective, decimals=4)

        tmp_zeros = np.zeros((1, nb_dispatchable), dtype=float)

        # wrap everything into the proper scipy form
        def target(actual_dispatchable):
            # define my real objective
            quad_ = (
                1e2
                * (
                    actual_dispatchable[gen_in_target]
                    - target_dispatch_redisp[gen_in_target]
                )
                ** 2
            )
            coeffs_quads = weights[gen_in_target] * quad_
            coeffs_quads_const = coeffs_quads.sum()
            coeffs_quads_const /= scale_objective  # scaling the function
            coeffs_quads_const += 1e-2 * (actual_dispatchable**2 * weights).sum()
            return coeffs_quads_const

        def jac(actual_dispatchable):
            res_jac = 1.0 * tmp_zeros
            res_jac[0, gen_in_target] = (
                1e2
                * 2.0
                * weights[gen_in_target]
                * (
                    actual_dispatchable[gen_in_target]
                    - target_dispatch_redisp[gen_in_target]
                )
            )
            res_jac /= scale_objective  # scaling the function
            res_jac += 2e-2 * actual_dispatchable * weights
            return res_jac

        mat_sum_ok = np.ones((1, nb_dispatchable))
        equality_const = LinearConstraint(
            mat_sum_ok, sum_target - self._tol_redisp, sum_target + self._tol_redisp
        )

        ineq_const = LinearConstraint(np.eye(nb_dispatchable), lb=val_min, ub=val_max)
        # objective function
        def f(init):
            this_res = minimize(
                target,
                init,
                method="SLSQP",
                constraints=[equality_const, ineq_const],
                options={
                    "eps": self._tol_redisp,
                    "ftol": self._tol_redisp,
                    "disp": False,
                },
                jac=jac,
            )
            return this_res

        # choose a good initial point (close to the solution)
        # the idea here is to chose a initial point that would be close to the
        # desired solution (split the (sum of the) dispatch to the available generators)
        x0 = 1.0 * target_dispatch_redisp
        can_adjust = x0 == 0.0
        if (can_adjust).any():
            init_sum = x0.sum()
            denom_adjust = (1.0 / weights[can_adjust]).sum()
            if denom_adjust <= 1e-2:
                # i don't want to divide by something too cloose to 0.
                denom_adjust = 1.0
            x0[can_adjust] = -init_sum / (weights[can_adjust] * denom_adjust)

        res = f(x0)
        if res.success:
            return res.x
        else:
            return None

    def _amount_curtailed(
        self, act: BaseAction, new_gen_p: np.ndarray
    ) -> Tuple[np.ndarray, float]:
        curt_vect = 1.0 * act.curtail
        curt_vect[curt_vect == -1.0] = 1.0
        limit_curtail = curt_vect * act.gen_pmax
        curtailed = np.maximum(new_gen_p - limit_curtail, 0.0)
        curtailed[~act.gen_renewable] = 0.0
        amount_curtail = curtailed.sum()
        new_gen_p_after_curtail = 1.0 * new_gen_p
        new_gen_p_after_curtail -= curtailed
        return new_gen_p_after_curtail, amount_curtail

    def _amount_storage(self, act: BaseAction) -> Tuple[float, np.ndarray]:
        storage_act = 1.0 * act.storage_p
        res = self.current_obs.storage_power_target.sum()
        current_charge = 1.0 * self.current_obs.storage_charge
        storage_power = np.zeros(act.n_storage)
        if np.all(np.abs(storage_act) <= self._tol_redisp):
            return -res, storage_power, current_charge
        coeff_p_to_E = (
            self.current_obs.delta_time / 60.0
        )  # obs.delta_time is in minutes

        # convert power (action to energy)
        storage_act_E = storage_act * coeff_p_to_E
        # take into account the efficiencies
        do_charge = storage_act_E < 0.0
        do_discharge = storage_act_E > 0.0
        storage_act_E[do_charge] /= act.storage_charging_efficiency[do_charge]
        storage_act_E[do_discharge] *= act.storage_discharging_efficiency[do_discharge]
        # make sure we don't go over / above Emin / Emax
        min_down_E = act.storage_Emin - current_charge
        min_up_E = act.storage_Emax - current_charge
        storage_act_E = np.minimum(storage_act_E, min_up_E)
        storage_act_E = np.maximum(storage_act_E, min_down_E)
        current_charge += storage_act_E
        # convert back to power (for the observation) the amount the grid got
        storage_power = storage_act_E / coeff_p_to_E
        storage_power[do_charge] *= act.storage_charging_efficiency[do_charge]
        storage_power[do_discharge] /= act.storage_discharging_efficiency[do_discharge]
        res += storage_power.sum()
        return -res, storage_power, current_charge

    def _fix_redisp_curtailment_storage(
        self, act: BaseAction, new_gen_p: np.ndarray
    ) -> Tuple[bool, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray,]:
        """This function emulates the "frequency control" of the
        environment.

        Its main goal is to ensure that the sum of injected power thanks to redispatching,
        storage units and curtailment sum to 0.

        It is a very rough simplification of what happens in the environment.
        """
        new_gen_p_after_curtail, amount_curtail = self._amount_curtailed(act, new_gen_p)
        amount_storage, storage_power, storage_charge = self._amount_storage(act)
        sum_target = amount_curtail - amount_storage  # TODO !

        target_dispatch = self.current_obs.target_dispatch + act.redispatch
        # if previous setpoint was say -2 and at this step I redispatch of
        # say + 4 then the real setpoint should be +2 (and not +4)
        new_vect_redisp = (act.redispatch != 0.0) & (
            self.current_obs.target_dispatch == 0.0
        )
        target_dispatch[new_vect_redisp] += self.current_obs.actual_dispatch[
            new_vect_redisp
        ]

        if abs(target_dispatch.sum() - sum_target) >= self._tol_redisp:
            adjust = self._adjust_controlable_gen(
                new_gen_p_after_curtail, target_dispatch, sum_target
            )
            if adjust is None:
                return True, None, None, None, None, None
            else:
                return (
                    True,
                    new_gen_p_after_curtail,
                    target_dispatch,
                    adjust,
                    storage_power,
                    storage_charge,
                )
        return False, None, None, None, None, None

    def predict(
        self,
        act: BaseAction,
        new_gen_p: np.ndarray = None,
        new_gen_v: np.ndarray = None,
        new_load_p: np.ndarray = None,
        new_load_q: np.ndarray = None,
        do_copy: bool = True,
    ) -> "Simulator":
        """Predict the state of the grid after a given action has been taken.

        Parameters
        ----------
        act : BaseAction
            The action you want to take
        new_gen_p : np.ndarray, optional
            the new production active setpoint, by default None
        new_gen_v : np.ndarray, optional
            the new production voltage setpoint, by default None
        new_load_p : np.ndarray, optional
            the new consumption active values, by default None
        new_load_q : np.ndarray, optional
            the new consumption reactive values, by default None
        do_copy : bool, optional
            Whether to make a copy or not, by default True
        
        Examples
        ---------
        
        A possible example is:
        
        .. code-block:: python
        
            import grid2op
            env_name = "l2rpn_case14_sandbox"  # or any other name
            env = grid2op.make(env_name)

            obs = env.reset()

            #### later in the code, for example in an Agent:

            simulator = obs.get_simulator()

            load_p_stressed = obs.load_p * 1.05
            gen_p_stressed = obs.gen_p * 1.05
            do_nothing = env.action_space()
            simulator_stressed = simulator.predict(act=do_nothing,
                                                new_gen_p=gen_p_stressed,
                                                new_load_p=load_p_stressed)
            if not simulator_stressed.converged:
                # the solver fails to find a solution for this action
                # you are likely to run into trouble if you use that...
                ...  # do something
            obs_stressed = simulator_stressed.current_obs
            
        Returns
        -------
        Simulator
            The new simulator representing the grid state after the simulation of the action.
            
        """
        # init the result
        if do_copy:
            res = self.copy()
        else:
            res = self
        this_act = act.copy()

        if new_gen_p is None:
            new_gen_p = 1.0 * self.current_obs.gen_p

        res.set_state(
            obs=None,
            new_gen_p=new_gen_p,
            new_gen_v=new_gen_v,
            new_load_p=new_load_p,
            new_load_q=new_load_q,
            do_powerflow=False,
        )

        # "fix" the action for the redispatching / curtailment / storage part
        (
            has_adjusted,
            new_gen_p_modif,
            target_dispatch,
            adjust,
            storage_power,
            storage_charge,
        ) = res._fix_redisp_curtailment_storage(this_act, new_gen_p)

        if has_adjusted:
            if target_dispatch is None:
                res._converged = False
                res.current_obs.set_game_over()
                res._error = InvalidRedispatching("")
                return res

            redisp_modif = np.zeros(self.current_obs.n_gen)
            redisp_modif[self.current_obs.gen_redispatchable] = adjust
            # adjust the proper things in the observation
            res.current_obs.target_dispatch = target_dispatch
            this_act.redispatch = redisp_modif
            res.current_obs.actual_dispatch[:] = redisp_modif
            this_act._dict_inj["prod_p"] = 1.0 * new_gen_p_modif
            this_act._modif_inj = True

            # TODO : curtail, curtailment_limit (in observation)
            res.current_obs.curtailment[:] = (
                new_gen_p - new_gen_p_modif
            ) / act.gen_pmax
            res.current_obs.curtailment_limit[:] = act.curtail
            res.current_obs.curtailment_limit_effective[:] = act.curtail
            res.current_obs.gen_p_before_curtail[:] = new_gen_p
            res.current_obs.storage_power[:] = storage_power
            res.current_obs.storage_charge[:] = storage_charge
        else:
            res.current_obs.storage_power[:] = 0.0
            res.current_obs.actual_dispatch[:] = 0.0

        # apply the action
        bk_act = res.backend.my_bk_act_class()
        bk_act += this_act
        res.backend.apply_action(bk_act)

        # run the powerflow
        res._do_powerflow()

        # update its observation
        res._update_obs()
        return res

    def close(self):
        """close the underlying backend"""
        if hasattr(self, "backend") and self.backend is not None:
            self.backend.close()
        self.backend = None
        self.current_obs = None
        self._converged = None
        self._error = None

    def __del__(self):
        self.close()



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np

from grid2op.Exceptions import Grid2OpException
from grid2op.Reward.baseReward import BaseReward
from grid2op.dtypes import dt_float


class EconomicReward(BaseReward):
    """
    This reward computes the marginal cost of the powergrid. As RL is about maximising a reward, while we want to
    minimize the cost, this class also ensures that:

    - the reward is positive if there is no game over, no error etc.
    - the reward is inversely proportional to the cost of the grid (the higher the reward, the lower the economic cost).

    Examples
    ---------
    You can use this reward in any environment with:

    .. code-block:

        import grid2op
        from grid2op.Reward import EconomicReward

        # then you create your environment with it:
        NAME_OF_THE_ENVIRONMENT = "rte_case14_realistic"
        env = grid2op.make(NAME_OF_THE_ENVIRONMENT,reward_class=EconomicReward)
        # and do a step with a "do nothing" action
        obs = env.reset()
        obs, reward, done, info = env.step(env.action_space())
        # the reward is computed with the EconomicReward class

    """

    def __init__(self, logger=None):
        BaseReward.__init__(self, logger=logger)
        self.reward_min = dt_float(0.0)
        self.reward_max = dt_float(1.0)
        self.worst_cost = None

    def initialize(self, env):
        if not env.redispatching_unit_commitment_availble:
            raise Grid2OpException(
                "Impossible to use the EconomicReward reward with an environment without generators"
                "cost. Please make sure env.redispatching_unit_commitment_availble is available."
            )
        self.worst_cost = dt_float((env.gen_cost_per_MW * env.gen_pmax).sum() * env.delta_time_seconds / 3600.0)

    def __call__(self, action, env, has_error, is_done, is_illegal, is_ambiguous):
        if has_error or is_illegal or is_ambiguous:
            res = self.reward_min
        else:
            # compute the cost of the grid
            res = dt_float((env.get_obs(_do_copy=False).prod_p * env.gen_cost_per_MW).sum() * env.delta_time_seconds / 3600.0)
            # we want to minimize the cost by maximizing the reward so let's take the opposite
            res *= dt_float(-1.0)
            # to be sure it's positive, add the highest possible cost
            res += self.worst_cost

        res = np.interp(
            res, [dt_float(0.0), self.worst_cost], [self.reward_min, self.reward_max]
        )
        return dt_float(res)



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
from grid2op.Action.baseAction import BaseAction


class CompleteAction(BaseAction):
    """
    Class representing the possibility to play everything.

    It cannot (and should not) be used by an Agent. Indeed, Agent actions are limited to :class:`PlayableAction`. This
    class is used by the chronics, the environment the opponent or the voltage controler for example.
    """

    def __init__(self):
        BaseAction.__init__(self)



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from grid2op.Exceptions.Grid2OpException import Grid2OpException


# Exception Runner is used twice, not possible on windows / macos due to the way multiprocessing works
class UsedRunnerError(Grid2OpException):
    """
    This exception indicate that runner (object of :class:`grid2op.Runner.Runner`) has already been used.

    This behaviour is not supported on windows / macos given the way the Multiprocessing package works (spawning
    a process where grid2op objects are made is not completly supported at the moment).

    The best solution is to recreate a runner, and then use this new one.
    """

    pass



#!/usr/bin/env python3

# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import warnings
import os
import unittest
import sys

from grid2op.main import main_cli as mainEntryPoint
from grid2op.Download.download import main as downloadEntryPoint

__LI_FILENAME_TESTS = [
    "test_Action.py",
    # "test_Action_iadd.py",
    "test_ActionProperties.py",
    "test_Observation.py",
    "test_AgentsFast.py",
    "test_RunnerFast.py",
    "test_attached_envs.py",
    # "test_GymConverter.py",  # requires gym
    # "test_Reward.py",
    # "test_issue_126.py",
    # "test_issue_131.py",
    # "test_issue_140.py",
    # "test_issue_146.py",
    # "test_issue_147.py",
    # # "test_issue_148.py",  # requires additional data
    # "test_issue_151.py",
    # "test_issue_153.py",
    # "test_issue_164.py",
]


def main():
    mainEntryPoint()


def download():
    downloadEntryPoint()


def replay():
    try:
        from grid2op.Episode.EpisodeReplay import main as replayEntryPoint

        replayEntryPoint()
    except ImportError as e:
        warn_msg = (
            "\nEpisode replay is missing an optional dependency\n"
            "Please run pip3 install grid2op[optional].\n The error was {}"
        )
        warnings.warn(warn_msg.format(e))


def testinstall():
    """
    Performs aperforms basic tests to make sure grid2op is properly installed and working.

    It's not because these tests pass that grid2op will be fully functional however.
    """
    test_loader = unittest.TestLoader()
    this_directory = os.path.abspath(os.path.dirname(__file__))
    test_suite = test_loader.discover(
        os.path.join(this_directory, "tests"), pattern=__LI_FILENAME_TESTS[0]
    )
    for file_name in __LI_FILENAME_TESTS[1:]:
        test_suite.addTest(
            test_loader.discover(
                os.path.join(this_directory, "tests"), pattern=file_name
            )
        )
    results = unittest.TextTestResult(stream=sys.stderr, descriptions=True, verbosity=1)
    test_suite.run(results)
    if results.wasSuccessful():
        sys.exit(0)
    else:
        for _, str_ in results.errors:
            print(str_)
            print("-------------------------\n")
        for _, str_ in results.failures:
            print(str_)
            print("-------------------------\n")
        raise RuntimeError("Test not successful !")



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np
from grid2op.Exceptions import IllegalAction
from grid2op.Rules.BaseRules import BaseRules


class PreventReconnection(BaseRules):
    """
    A subclass is used to check that an action will not attempt to reconnect a powerlines disconnected because of
    an overflow, or to check that 2 actions acting on the same powerline are distant from the right number of timesteps
    (see :attr:`grid2op.Parameters.Parameters.NB_TIMESTEP_LINE_STATUS_REMODIF`) or if two topological modification
    of the same substation are too close in time
    (see :attr:`grid2op.Parameters.Parameters.NB_TIMESTEP_TOPOLOGY_REMODIF`)

    """

    def __call__(self, action, env):
        """
        This function check only that the action doesn't attempt to reconnect  a powerline that has been disconnected
        due to an overflow.

        See :func:`BaseRules.__call__` for a definition of the parameters of this function.

        """
        # at first iteration, env.current_obs is None...
        # TODO this is used inside the environment (for step) inside LookParam and here
        # this could be computed only once, and fed to this instead
        powerline_status = env.get_current_line_status()

        aff_lines, aff_subs = action.get_topological_impact(powerline_status)
        if (env._times_before_line_status_actionable[aff_lines] > 0).any():
            # i tried to act on a powerline too shortly after a previous action
            # or shut down due to an overflow or opponent or hazards or maintenance
            ids = np.where((env._times_before_line_status_actionable > 0) & aff_lines)[
                0
            ]
            return False, IllegalAction(
                "Powerline with ids {} have been modified illegally (cooldown of {})".format(
                    ids, env._times_before_line_status_actionable[ids]
                )
            )

        if (env._times_before_topology_actionable[aff_subs] > 0).any():
            # I tried to act on a topology too shortly after a previous action
            ids = np.where((env._times_before_topology_actionable > 0) & aff_subs)[0]
            return False, IllegalAction(
                "Substation with ids {} have been modified illegally (cooldown of {})".format(
                    ids, env._times_before_topology_actionable[ids]
                )
            )

        return True, None



__all__ = ["EpisodeData"]

from grid2op.Episode.EpisodeData import EpisodeData

# Try to import optional module
try:
    from grid2op.Episode.EpisodeReplay import EpisodeReplay

    __all__.append("EpisodeReplay")
except ImportError:
    pass  # Silent fail for optional dependencies



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from grid2op.Rules.BaseRules import BaseRules


class AlwaysLegal(BaseRules):
    """
    This subclass doesn't implement any rules regarding the legality of the actions. All actions are legal.

    """

    def __call__(self, action, env):
        """
        All actions being legal, this returns always true.
        See :func:`BaseRules.__call__` for a definition of the parameters of this function.

        """
        return True, None



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

"""
TODO documentation of this function!

"""
import os

import argparse

from grid2op.Observation import CompleteObservation
from grid2op.Chronics import Multifolder
from grid2op.Reward import FlatReward
from grid2op.Agent import DoNothingAgent
from grid2op.Backend import PandaPowerBackend
from grid2op.Rules import AlwaysLegal
from grid2op.Runner import Runner


def main_run(
    path_casefile=None,
    path_chronics=None,
    path_parameters=None,
    chronics_class=Multifolder,
    backend_class=PandaPowerBackend,
    agent_class=DoNothingAgent,
    reward_class=FlatReward,
    observation_class=CompleteObservation,
    legalAct_class=AlwaysLegal,
    nb_episode=3,
    nb_process=1,
    path_save=None,
    names_chronics_to_backend=None,
    gridStateclass_kwargs={},
):
    init_grid_path = os.path.abspath(path_casefile)

    path_chron = os.path.abspath(path_chronics)

    parameters_path = path_parameters

    runner = Runner(
        init_grid_path=init_grid_path,
        path_chron=path_chron,
        parameters_path=parameters_path,
        names_chronics_to_backend=names_chronics_to_backend,
        gridStateclass=chronics_class,
        gridStateclass_kwargs=gridStateclass_kwargs,
        backendClass=backend_class,
        rewardClass=reward_class,
        agentClass=agent_class,
        observationClass=observation_class,
        legalActClass=legalAct_class,
    )

    res = runner.run(nb_episode=nb_episode, nb_process=nb_process, path_save=path_save)
    return res


def cli_main():
    parser = argparse.ArgumentParser(
        description='Launch the evaluation of the Grid2Op ("Grid To Operate") code.'
    )
    parser.add_argument(
        "--path_save",
        default=None,
        help="The path where the log of the experience will be stored (default: None -> nothing stored)",
    )
    parser.add_argument(
        "--nb_process",
        type=int,
        default=1,
        help="The number of process used for each evaluation (note that if nb_process > nb_episode then nb_episode is used.",
    )
    parser.add_argument(
        "--nb_episode",
        type=int,
        default=3,
        help="The number of episode to play (default 3)",
    )
    parser.add_argument(
        "--path_casefile",
        type=str,
        required=True,
        help="Path where the case file is located (casefile is the file describing the powergrid)",
    )
    parser.add_argument(
        "--path_chronics",
        type=str,
        required=True,
        help="Path where the chronics (temporal variation of loads and production usually are located)",
    )
    parser.add_argument(
        "--path_parameters",
        default=None,
        help="Path where the _parameters of the game are stored",
    )

    args = parser.parse_args()
    return args


def main_cli(args=None):
    if args is None:
        args = cli_main()

    if args.path_save is not None:
        path_save = str(args.path_save)
    else:
        path_save = None
    if args.path_parameters is not None:
        path_parameter = str(args.path_parameters)
    else:
        path_parameter = None

    names_chronics_to_backend = None

    # actually performing the run
    msg_ = 'Running Grid2Op:\n\t- on case file at "{case_file}"\n\t- with data located at "{data}"'
    msg_ += "\n\t- using {process} process(es)\n\t- for {nb_episode} episodes"
    if args.path_save is None:
        msg_ += "\n\t- results will not be saved"
    else:
        msg_ += '\n\t- results will be saved in "{}"'.format(args.path_save)
    print(
        msg_.format(
            case_file=args.path_casefile,
            data=args.path_chronics,
            process=args.nb_process,
            nb_episode=args.nb_episode,
        )
    )
    res = main_run(
        path_save=path_save,
        nb_process=args.nb_process,
        nb_episode=args.nb_episode,
        path_casefile=args.path_casefile,
        path_chronics=args.path_chronics,
        path_parameters=path_parameter,
        names_chronics_to_backend=names_chronics_to_backend,
    )
    print("The results are:")
    for chron_name, _, cum_reward, nb_time_step, max_ts in res:
        msg_tmp = "\tFor chronics located at {}\n".format(chron_name)
        msg_tmp += "\t\t - cumulative reward: {:.2f}\n".format(cum_reward)
        msg_tmp += "\t\t - number of time steps completed: {:.0f} / {:.0f}".format(
            nb_time_step, max_ts
        )
        print(msg_tmp)


if __name__ == "__main__":
    args = cli_main()
    main_cli(args)



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.


class GymObservationSpace:
    def __init__(self, env):
        raise RuntimeError(
            'The "GymObservationSpace" has been moved to "grid2op.gym" module instead.\n'
            'Note to update: use "from grid2op.gym_compat import GymObservationSpace"'
        )


class GymActionSpace:
    def __init__(self, action_space):
        raise RuntimeError(
            'The "GymActionSpace" has been moved to "grid2op.gym" module instead.\n'
            'Note to update: use "from grid2op.gym_compat import GymActionSpace"'
        )



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from abc import abstractmethod

from grid2op.Converter import Converter
from grid2op.Exceptions import Grid2OpException
from grid2op.Agent.baseAgent import BaseAgent


class AgentWithConverter(BaseAgent):
    """
    Compared to a regular BaseAgent, these types of Agents are able to deal with a different representation of
    :class:`grid2op.Action.BaseAction` and :class:`grid2op.Observation.BaseObservation`.

    As any other Agents, AgentWithConverter will implement the :func:`BaseAgent.act` method. But for them, it's slightly
    different.

    They receive in this method an observation, as an object (ie an instance of
    :class:`grid2op.Observation.BaseObservation`). This
    object can then be converted to any other object with the method :func:`AgentWithConverter.convert_obs`.

    Then, this `transformed_observation` is pass to the method :func:`AgentWithConverter.my_act` that is supposed
    to be defined for each agents. This function outputs an `encoded_act` which can be whatever you want to be.

    Finally, the `encoded_act` is decoded into a proper action, object of class :class:`grid2op.Action.BaseAction`,
    thanks to the method :func:`AgentWithConverter.convert_act`.

    This allows, for example, to represent actions as integers to train more easily standard discrete control algorithm
    used to solve atari games for example.

    **NB** It is possible to define :func:`AgentWithConverter.convert_obs` and :func:`AgentWithConverter.convert_act`
     or to define a :class:`grid2op.Converters.Converter` and feed it to the `action_space_converter` parameters
     used to initialise the class. The second option is preferred, as the :attr:`AgentWithConverter.action_space`
     will then directly be this converter. Such an BaseAgent will really behave as if the actions are encoded the way he
     wants.

    Examples
    --------
    For example, imagine an BaseAgent uses a neural networks to take its decision.

    Suppose also that, after some
    features engineering, it's best for the neural network to use only the load active values
    (:attr:`grid2op.Observation.BaseObservation.load_p`) and the sum of the
    relative flows (:attr:`grid2op.Observation.BaseObservation.rho`) with the active flow
    (:attr:`grid2op.Observation.BaseObservation.p_or`) [**NB** that agent would not make sense a priori, but who knows]

    Suppose that this neural network can be accessed with a class `AwesomeNN` (not available...) that can
    predict some actions. It can be loaded with the "load" method and make predictions with the
    "predict" method.

    For the sake of the examples, we will suppose that this agent only predicts powerline status (so 0 or 1) that
    are represented as vector. So we need to take extra care to convert this vector from a numpy array to a valid
    action.

    This is done below:


    .. code-block:: python

        import grid2op
        import AwesomeNN # this does not exists!
        # create a simple environment
        env = grid2op.make()

        # define the class above
        class AgentCustomObservation(AgentWithConverter):
            def __init__(self, action_space, path):
                AgentWithConverter.__init__(self, action_space)
                self.my_neural_network = AwesomeNN()
                self.my_neural_networl.load(path)

            def convert_obs(self, observation):
                # convert the observation
                return np.concatenate((observation.load_p, observation.rho + observation.p_or))

            def convert_act(self, encoded_act):
                # convert back the action, output from the NN "self.my_neural_network"
                # to a valid action
                act = self.action_space({"set_status": encoded_act})

            def my_act(self, transformed_observation, reward, done=False):
                act_predicted = self.my_neural_network(transformed_observation)
                return act_predicted


        # make the agent that behaves as expected.
        my_agent = AgentCustomObservation(action_space=env.action_space, path=".")

        # this agent is perfectly working :-) You can use it as any other agents.


    Attributes
    ----------
    action_space_converter: :class:`grid2op.Converters.Converter`
        The converter that is used to represents the BaseAgent action space. Might be set to ``None`` if not initialized

    init_action_space: :class:`grid2op.Action.ActionSpace`
        The initial action space. This corresponds to the action space of the :class:`grid2op.Environment.Environment`.

    action_space: :class:`grid2op.Converters.ActionSpace`
        If a converter is used, then this action space represents is this converter. The agent will behave as if
        the action space is directly encoded the way it wants.

    """

    def __init__(self, action_space, action_space_converter=None, **kwargs_converter):
        self.action_space_converter = action_space_converter
        self.init_action_space = action_space

        if action_space_converter is None:
            BaseAgent.__init__(self, action_space)
        else:
            if isinstance(action_space_converter, type):
                if issubclass(action_space_converter, Converter):
                    action_space_converter_this_env_class = (
                        action_space_converter.init_grid(action_space)
                    )
                    this_action_space = action_space_converter_this_env_class(
                        action_space
                    )
                    BaseAgent.__init__(self, this_action_space)
                else:
                    raise Grid2OpException(
                        "Impossible to make an BaseAgent with a converter of type {}. "
                        "Please use a converter deriving from grid2op.ActionSpaceConverter.Converter."
                        "".format(action_space_converter)
                    )
            elif isinstance(action_space_converter, Converter):
                if isinstance(
                    action_space_converter._template_act,
                    self.init_action_space.actionClass,
                ):
                    BaseAgent.__init__(self, action_space_converter)
                else:
                    raise Grid2OpException(
                        "Impossible to make an BaseAgent with the provided converter of type {}. "
                        "It doesn't use the same type of action as the BaseAgent's action space."
                        "".format(action_space_converter)
                    )
            else:
                raise Grid2OpException(
                    'You try to initialize and BaseAgent with an invalid converter "{}". It must'
                    'either be a type deriving from "Converter", or an instance of a class'
                    "deriving from it."
                    "".format(action_space_converter)
                )

            self.action_space.init_converter(**kwargs_converter)

    def convert_obs(self, observation):
        """
        This function convert the observation, that is an object of class :class:`grid2op.Observation.BaseObservation`
        into a representation understandable by the BaseAgent.

        For example, and agent could only want to look at the relative flows
        :attr:`grid2op.Observation.BaseObservation.rho`
        to take his decision. This is possible by  overloading this method.

        This method can also be used to scale the observation such that each compononents has mean 0 and variance 1
        for example.

        Parameters
        ----------
        observation: :class:`grid2op.Observation.Observation`
            Initial observation received by the agent in the :func:`BaseAgent.act` method.

        Returns
        -------
        res: ``object``
            Anything that will be used by the BaseAgent to take decisions.

        """
        return self.action_space.convert_obs(observation)

    def convert_act(self, encoded_act):
        """
        This function will convert an "ecnoded action" that be of any types, to a valid action that can be ingested
        by the environment.

        Parameters
        ----------
        encoded_act: ``object``
            Anything that represents an action.

        Returns
        -------
        act: :grid2op.BaseAction.BaseAction`
            A valid actions, represented as a class, that corresponds to the encoded action given as input.

        """
        res = self.action_space.convert_act(encoded_act)
        return res

    def act(self, observation, reward, done=False):
        """
        Standard method of an :class:`BaseAgent`. There is no need to overload this function.

        Parameters
        ----------
        observation: :class:`grid2op.Observation.Observation`
            The current observation of the :class:`grid2op.Environment.Environment`

        reward: ``float``
            The current reward. This is the reward obtained by the previous action

        done: ``bool``
            Whether the episode has ended or not. Used to maintain gym compatibility

        Returns
        -------
        res: :class:`grid2op.Action.Action`
            The action chosen by the bot / controler / agent.

        """
        transformed_observation = self.convert_obs(observation)
        encoded_act = self.my_act(transformed_observation, reward, done)
        return self.convert_act(encoded_act)

    def seed(self, seed):
        """
        Seed the agent AND the associated converter if it needs to be seeded.

        See a more detailed explanation in :func:`BaseAgent.seed` for more information about seeding.
        """
        super().seed(seed)
        if self.action_space_converter is not None:
            self.action_space.seed(seed)

    @abstractmethod
    def my_act(self, transformed_observation, reward, done=False):
        """
        This method should be override if this class is used. It is an "abstract" method.

        If someone wants to make a agent that handles different kinds of actions an observation.

        Parameters
        ----------
        transformed_observation: ``object``
            Anything that will be used to create an action. This is the results to the call of
            :func:`AgentWithConverter.convert_obs`. This is likely a numpy array.

        reward: ``float``
            The current reward. This is the reward obtained by the previous action

        done: ``bool``
            Whether the episode has ended or not. Used to maintain gym compatibility

        Returns
        -------
        res: ``object``
            A representation of an action in any possible format. This action will then be ingested and formatted into
            a valid action with the :func:`AgentWithConverter.convert_act` method.

        """
        transformed_action = None
        return transformed_action



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import copy
import warnings
# from gym.spaces import Discrete

from grid2op.Exceptions import Grid2OpException
from grid2op.Action import ActionSpace
from grid2op.Converter import IdToAct

from grid2op.gym_compat.utils import (ALL_ATTR_FOR_DISCRETE,
                                      ATTR_DISCRETE,
                                      GYM_AVAILABLE,
                                      GYMNASIUM_AVAILABLE)

# TODO test that it works normally
# TODO test the casting in dt_int or dt_float depending on the data
# TODO test the scaling
# TODO doc
# TODO test the function part


class __AuxDiscreteActSpace:
    """
    TODO the documentation of this class is in progress.

    This class allows to convert a grid2op action space into a gym "Discrete". This means that the action are
    labeled, and instead of describing the action itself, you provide only its ID.
    
    Let's take an example of line disconnection. In the "standard" gym representation you need to:
    
    .. code-block:: python

        import grid2op
        import numpy as np
        from grid2op.gym_compat import GymEnv
        
        env_name = "l2rpn_case14_sandbox"  # or any other name
        env = grid2op.make(env_name)
        gym_env = GymEnv(env)

        # now do an action
        gym_act = {}
        gym_act["set_bus"]  = np.zeros(env.n_line, dtype=np.int)
        l_id = ... # the line you want to disconnect
        gym_act["set_bus"][l_id] = -1
        obs, reward, done, truncated, info = gym_env.step(gym_act)
        
    This has the advantage to be as close as possible to raw grid2op. But the main drawback is that
    most of RL framework are not able to do this kind of modification easily. For discrete actions,
    what is often do is:
    
    1) enumerate all possible actions (say you have n different actions)
    2) assign a unique id to all actions (say from 0 to n-1)
    3) have a "policy" output a vector of size n with each component
       representing an action (eg `vect[42]` represents the score the policy assign to action `42`)    
       
    Instead of having everyone doing the modifications "on its own" we developed the :class:`DiscreteActSpace`
    that does exactly this, in a single line of code:
    
    .. code-block:: python

        import grid2op
        import numpy as np
        from grid2op.gym_compat import GymEnv, DiscreteActSpace

        env_name = "l2rpn_case14_sandbox"  # or any other name
        env = grid2op.make(env_name)
        gym_env = GymEnv(env)
        gym_env.action_space = DiscreteActSpace(env.action_space,
                                                attr_to_keep=["set_bus", 
                                                              "set_line_status",
                                                              # or anything else
                                                             ]
                                                )

        # do action with ID 42
        gym_act = 42
        obs, reward, done, truncated, info = gym_env.step(gym_act)    
        # to know what the action did, you can
        # print(gym_env.action_space.from_gym(gym_act))
    
    
    It is related to the :class:`MultiDiscreteActSpace` but compared to this other representation, it
    does not allow to do "multiple actions". Typically, if you use the snippets below:

    .. code-block:: python

        import grid2op
        env_name = "l2rpn_case14_sandbox"  # or any other name
        env = grid2op.make(env_name)

        from grid2op.gym_compat import GymEnv, MultiDiscreteActSpace, DiscreteActSpace
        gym_env1 = GymEnv(env)
        gym_env2 = GymEnv(env)

        gym_env1.action_space = MultiDiscreteActSpace(env.action_space,
                                                      attr_to_keep=['redispatch', "curtail", "one_sub_set"])
        gym_env2.action_space = DiscreteActSpace(env.action_space,
                                                 attr_to_keep=['redispatch', "curtail", "set_bus"])

    Then at each step, `gym_env1` will allow to perform a redispatching action (on any number of generators),
    a curtailment
    action (on any number of generators) __**AND**__ changing the topology at one substation. But at each
    steps, the agent should predicts lots of "number".

    On the other hand, at each step, the agent for `gym_env2` will have to predict a single integer (which is
    usually the case in most RL environment) but it this action will affect redispatching on a single generator,
    perform curtailment on a single generator __**OR**__  changing the topology at one substation. But at each
    steps, the agent should predicts only one "number".

    The action set is then largely  constrained compared to the :class:`MultiDiscreteActSpace`

    .. note::
        This class is really closely related to the :class:`grid2op.Converter.IdToAct`. It basically "maps"
        this "IdToAct" into a type of gym space, which, in this case, will be a `Discrete` one.
        
    .. note::
        By default, the "do nothing" action is encoded by the integer '0'.

    .. warning::
        Depending on the presence absence of gymnasium and gym packages this class might behave differently.
        
        In grid2op we tried to maintain compatibility both with gymnasium (newest) and gym (legacy, 
        no more maintained) RL packages. The behaviour is the following:
        
        - :class:`DiscreteActSpace` will inherit from gymnasium if it's installed 
          (in this case it will be :class:`DiscreteActSpaceGymnasium`), otherwise it will
          inherit from gym (and will be exactly :class:`DiscreteActSpaceLegacyGym`)
        - :class:`DiscreteActSpaceGymnasium` will inherit from gymnasium if it's available and never from
          from gym
        - :class:`DiscreteActSpaceLegacyGym` will inherit from gym if it's available and never from
          from gymnasium
        
        See :ref:`gymnasium_gym` for more information
        
    Examples
    --------

    We recommend to use it like:

    .. code-block:: python

        import grid2op
        env_name = "l2rpn_case14_sandbox"  # or any other name
        env = grid2op.make(env_name)

        from grid2op.gym_compat import GymEnv, DiscreteActSpace
        gym_env = GymEnv(env)

        gym_env.observation_space = DiscreteActSpace(env.observation_space,
                                                     attr_to_keep=['redispatch', "curtail", "set_bus"])

    The possible attribute you can provide in the "attr_to_keep" are:

    - "set_line_status"
    - "set_line_status_simple" (grid2op >= 1.6.6) : set line status adds 5 actions per powerlines: 
      
      1) disconnect it
      2) connect origin side to busbar 1 and extermity side to busbar 1
      3) connect origin side to busbar 1 and extermity side to busbar 2
      4) connect origin side to busbar 2 and extermity side to busbar 1
      5) connect origin side to busbar 2 and extermity side to busbar 2
      
      This is "over complex" for most use case where you just want to "connect it"
      or "disconnect it". If you want the simplest version, just use "set_line_status_simple".
    - "change_line_status"
    - "set_bus": corresponds to changing the topology using the "set_bus" (equivalent to the
      "one_sub_set" keyword in the "attr_to_keep" of the :class:`MultiDiscreteActSpace`)
    - "change_bus": corresponds to changing the topology using the "change_bus" (equivalent to the
      "one_sub_change" keyword in the "attr_to_keep" of the :class:`MultiDiscreteActSpace`)
    - "redispatch"
    - "set_storage"
    - "curtail"
    - "curtail_mw" (same effect as "curtail")

    If you do not want (each time) to build all the actions from the action space, but would rather
    save the actions you find the most interesting and then reload them, you can, for example:

    .. code-block:: python

        import grid2op
        from grid2op.gym_compat import GymEnv, DiscreteActSpace
        env_name = "l2rpn_case14_sandbox"  # or any other name
        env = grid2op.make(env_name)

        gym_env = GymEnv(env)
        action_list = ... # a list of action, that can be processed by
        # IdToAct.init_converter (all_actions): see
        # https://grid2op.readthedocs.io/en/latest/converter.html#grid2op.Converter.IdToAct.init_converter
        gym_env.observation_space = DiscreteActSpace(env.observation_space,
                                                     action_list=action_list)

    .. note::

        This last version (providing explicitly the actions you want to keep and their ID)
        is much (much) safer and reproducible. Indeed, the
        actions usable by your agent will be the same (and in the same order)
        regardless of the grid2op version, of the person using it, of pretty
        much everything.

        It might not be consistent (between different grid2op versions)
        if the actions are built from scratch (for example, depending on the
        grid2op version other types of actions can be made, such as curtailment,
        or actions on storage units) like it's the case with the key-words
        (*eg* "set_bus") you pass as argument in the `attr_to_keep`

    """

    def __init__(
        self,
        grid2op_action_space,
        attr_to_keep=ALL_ATTR_FOR_DISCRETE,
        nb_bins=None,
        action_list=None,
    ):

        if not isinstance(grid2op_action_space, ActionSpace):
            raise Grid2OpException(
                f"Impossible to create a BoxGymActSpace without providing a "
                f"grid2op action_space. You provided {type(grid2op_action_space)}"
                f'as the "grid2op_action_space" attribute.'
            )

        if nb_bins is None:
            nb_bins = {"redispatch": 7, "set_storage": 7, "curtail": 7}

        if "raise_alert" in attr_to_keep or "raise_alarm" in attr_to_keep:
            raise Grid2OpException("This converter cannot be use to raise alarm or raise alert. "
                                   "Please use the MultiDiscreteActSpace space for this purpose.")
        
        act_sp = grid2op_action_space
        self.action_space = copy.deepcopy(act_sp)

        if attr_to_keep == ALL_ATTR_FOR_DISCRETE:
            # by default, i remove all the attributes that are not supported by the action type
            # i do not do that if the user specified specific attributes to keep. This is his responsibility in
            # in this case
            attr_to_keep = {
                el for el in attr_to_keep if grid2op_action_space.supports_type(el)
            }
        else:
            if action_list is not None:
                raise Grid2OpException(
                    "Impossible to specify a list of attributes "
                    "to keep (argument attr_to_keep) AND a list of "
                    "action to use (argument action_list)."
                )
        for el in attr_to_keep:
            if el not in ATTR_DISCRETE and action_list is None:
                warnings.warn(
                    f'The class "DiscreteActSpace" should mainly be used to consider only discrete '
                    f"actions (eg. set_line_status, set_bus or change_bus). Though it is possible to use "
                    f'"{el}" when building it, be aware that this continuous action will be treated '
                    f"as discrete by splitting it into bins. "
                    f'Consider using the "BoxGymActSpace" for these attributes.'
                )

        self._attr_to_keep = sorted(attr_to_keep)
        self._nb_bins = nb_bins

        self.dict_properties = {
            "set_line_status": act_sp.get_all_unitary_line_set,
            "change_line_status": act_sp.get_all_unitary_line_change,
            "set_bus": act_sp.get_all_unitary_topologies_set,
            "change_bus": act_sp.get_all_unitary_topologies_change,
            "redispatch": act_sp.get_all_unitary_redispatch,
            "set_storage": act_sp.get_all_unitary_storage,
            "curtail": act_sp.get_all_unitary_curtail,
            "curtail_mw": act_sp.get_all_unitary_curtail,
            # "raise_alarm": act_sp.get_all_unitary_alarm,
            # "raise_alert": act_sp.get_all_unitary_alert,
            "set_line_status_simple": act_sp.get_all_unitary_line_set_simple,
        }

        if action_list is None:
            self.converter = None
            n_act = self._get_info()
        else:
            self.converter = IdToAct(self.action_space)
            self.converter.init_converter(all_actions=action_list)
            n_act = self.converter.n

        # initialize the base container
        type(self)._DiscreteType.__init__(self, n=n_act)

    def _get_info(self):
        converter = IdToAct(self.action_space)
        li_act = [self.action_space()]
        for attr_nm in self._attr_to_keep:
            if attr_nm in self.dict_properties:
                if attr_nm not in self._nb_bins:
                    li_act += self.dict_properties[attr_nm](self.action_space)
                else:
                    if attr_nm == "curtail" or attr_nm == "curtail_mw":
                        li_act += self.dict_properties[attr_nm](
                            self.action_space, num_bin=self._nb_bins[attr_nm]
                        )
                    else:
                        li_act += self.dict_properties[attr_nm](
                            self.action_space,
                            num_down=self._nb_bins[attr_nm],
                            num_up=self._nb_bins[attr_nm],
                        )
            else:
                li_keys = "\n\t- ".join(sorted(list(self.dict_properties.keys())))
                raise RuntimeError(
                    f'Unknown action attributes "{attr_nm}". Supported attributes are: '
                    f"\n\t- {li_keys}"
                )

        converter.init_converter(li_act)
        self.converter = converter
        return self.converter.n

    def from_gym(self, gym_act):
        """
        This is the function that is called to transform a gym action (in this case a numpy array!)
        sent by the agent
        and convert it to a grid2op action that will be sent to the underlying grid2op environment.

        Parameters
        ----------
        gym_act: ``int``
            the gym action (a single integer for this action space)

        Returns
        -------
        grid2op_act: :class:`grid2op.Action.BaseAction`
            The corresponding grid2op action.

        """
        res = self.converter.all_actions[int(gym_act)]
        return res

    def close(self):
        pass


if GYM_AVAILABLE:
    from gym.spaces import Discrete as LegGymDiscrete
    from grid2op.gym_compat.box_gym_actspace import BoxLegacyGymActSpace
    from grid2op.gym_compat.continuous_to_discrete import ContinuousToDiscreteConverterLegacyGym
    DiscreteActSpaceLegacyGym = type("DiscreteActSpaceLegacyGym",
                                     (__AuxDiscreteActSpace, LegGymDiscrete, ),
                                     {"_gymnasium": False,
                                      "_DiscreteType": LegGymDiscrete,
                                      "_BoxGymActSpaceType": BoxLegacyGymActSpace,
                                      "_ContinuousToDiscreteConverterType": ContinuousToDiscreteConverterLegacyGym,
                                      "__module__": __name__})
    DiscreteActSpaceLegacyGym.__doc__ = __AuxDiscreteActSpace.__doc__
    DiscreteActSpace = DiscreteActSpaceLegacyGym
    DiscreteActSpace.__doc__ = __AuxDiscreteActSpace.__doc__
        

if GYMNASIUM_AVAILABLE:
    from gymnasium.spaces import Discrete
    from grid2op.gym_compat.box_gym_actspace import BoxGymnasiumActSpace
    from grid2op.gym_compat.continuous_to_discrete import  ContinuousToDiscreteConverterGymnasium
    DiscreteActSpaceGymnasium = type("MultiDiscreteActSpaceGymnasium",
                                     (__AuxDiscreteActSpace, Discrete, ),
                                     {"_gymnasium": True,
                                      "_DiscreteType": Discrete,
                                      "_BoxGymActSpaceType": BoxGymnasiumActSpace,
                                      "_ContinuousToDiscreteConverterType": ContinuousToDiscreteConverterGymnasium,
                                      "__module__": __name__})
    DiscreteActSpaceGymnasium.__doc__ = __AuxDiscreteActSpace.__doc__
    DiscreteActSpace = DiscreteActSpaceGymnasium
    DiscreteActSpace.__doc__ = __AuxDiscreteActSpace.__doc__



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np


class PlotUtil:
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

    This is a wrapper that contains utilities to draw the information on the plots more easily.
    """

    @staticmethod
    def format_value_unit(value, unit):
        if isinstance(value, float):
            return "{:.2f} {}".format(value, unit)
        elif isinstance(value, int):
            return "{:d} {}".format(value, unit)
        else:
            return "{} {}".format(value, unit)

    @staticmethod
    def middle_from_points(x1, y1, x2, y2):
        return (x1 + x2) * 0.5, (y1 + y2) * 0.5

    @staticmethod
    def vec_from_points(x1, y1, x2, y2):
        return x2 - x1, y2 - y1

    @staticmethod
    def mag_from_points(x1, y1, x2, y2):
        x, y = PlotUtil.vec_from_points(x1, y1, x2, y2)
        return np.linalg.norm([x, y])

    @staticmethod
    def norm_from_points(x1, y1, x2, y2):
        (
            x,
            y,
        ) = PlotUtil.vec_from_points(x1, y1, x2, y2)
        return PlotUtil.norm_from_vec(x, y)

    @staticmethod
    def norm_from_vec(x, y):
        n = np.linalg.norm([x, y])
        return x / n, y / n

    @staticmethod
    def orth_from_points(x1, y1, x2, y2):
        x, y = PlotUtil.vec_from_points(x1, y1, x2, y2)
        return -y, x

    @staticmethod
    def orth_norm_from_points(x1, y1, x2, y2):
        x, y = PlotUtil.vec_from_points(x1, y1, x2, y2)
        xn, yn = PlotUtil.norm_from_vec(x, y)
        return -yn, xn



from grid2op.Action import PlayableAction
from grid2op.Reward import RedispReward
from grid2op.Rules import DefaultRules
from grid2op.Chronics import Multifolder
from grid2op.Chronics import GridStateFromFileWithForecasts
from grid2op.Backend import PandaPowerBackend

config = {
    "backend": PandaPowerBackend,
    "action_class": PlayableAction,
    "observation_class": None,
    "reward_class": RedispReward,
    "gamerules_class": DefaultRules,
    "chronics_class": Multifolder,
    "grid_value_class": GridStateFromFileWithForecasts,
    "volagecontroler_class": None,
    "thermal_limits": [
        541.0,
        450.0,
        375.0,
        636.0,
        175.0,
        285.0,
        335.0,
        657.0,
        496.0,
        827.0,
        442.0,
        641.0,
        840.0,
        156.0,
        664.0,
        235.0,
        119.0,
        179.0,
        1986.0,
        1572.0,
    ],
    "names_chronics_to_grid": None
}



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
import numpy as np
from grid2op.Agent.greedyAgent import GreedyAgent


class RecoPowerlineAgent(GreedyAgent):
    """
    This is a :class:`GreedyAgent` example, which will attempt to reconnect powerlines: for each disconnected powerline
    that can be reconnected, it will simulate the effect of reconnecting it. And reconnect the one that lead to the
    highest simulated reward.

    """

    def __init__(self, action_space):
        GreedyAgent.__init__(self, action_space)

    def _get_tested_action(self, observation):
        res = [self.action_space({})]  # add the do nothing
        line_stat_s = observation.line_status
        cooldown = observation.time_before_cooldown_line
        can_be_reco = ~line_stat_s & (cooldown == 0)
        if can_be_reco.any():
            res = [
                self.action_space({"set_line_status": [(id_, +1)]})
                for id_ in np.where(can_be_reco)[0]
            ]
        return res



# Copyright (c) 2019-2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np

from typing import Optional
from grid2op.Reward.baseReward import BaseReward
from grid2op.dtypes import dt_float, dt_bool, dt_int


class AlertReward(BaseReward):
    """
    .. note::
        DOC IN PROGRESS !
    
    This reward is based on the "alert feature" where the agent is asked to send information about potential line overload issue
    on the grid after unpredictable powerline disconnection (attack of the opponent). The alerts are assessed once per attack.


    This rewards is computed as followed:
    
    - if an attack occurs and the agent survives `env.parameters.ALERT_TIME_WINDOW` steps then:
      - if the agent sent an alert BEFORE the attack, reward returns `reward_min_no_blackout` (-1 by default)
      - if the agent did not sent an alert BEFORE the attack, reward returns `reward_max_no_blackout` (1 by default)
    - if an attack occurs and the agent "games over" withing `env.parameters.ALERT_TIME_WINDOW` steps then:
      - if the agent sent an alert BEFORE the attack, reward returns `reward_max_blackout` (2 by default)
      - if the agent did not sent an alert BEFORE the attack, reward returns `reward_min_blackout` (-10 by default)
    - whatever the attacks / no attacks / alert / no alert, if the scenario is completed until the end, 
      then agent receive `reward_end_episode_bonus` (1 by default)

    In all other cases, including but not limited to:
    
    - agent games over but there has been no attack within the previous `env.parameters.ALERT_TIME_WINDOW` (12) steps
    - there is no attack 
    
    The reward outputs 0.
    
    This is then a "delayed reward": you receive the reward (in general) `env.parameters.ALERT_TIME_WINDOW` after
    having sent the alert.
    
    This is also a "sparse reward": in the vast majority of cases it's 0. It is only non zero in case of blackout (at
    most once per episode) and each time an attack occurs (and in general there is relatively few attacks)
    
    TODO explain a bit more in the "multi lines attacked"
    
    .. seealso:: :ref:`grid2op-alert-module` section of the doc for more information
    
    Examples
    ---------
    You can use this reward in any environment with:

    .. code-block:

        import grid2op
        from grid2op.Reward import AlertReward

        # then you create your environment with it:
        # at time of writing, the only env supporting it is "l2rpn_idf_2023"
        NAME_OF_THE_ENVIRONMENT = "l2rpn_idf_2023"  
        
        env = grid2op.make(NAME_OF_THE_ENVIRONMENT, reward_class=AlertReward)
        # and do a step with a "do nothing" action
        obs = env.reset()
        obs, reward, done, info = env.step(env.action_space())
        # the reward is computed with the AlertReward class


    """

    def __init__(self,
                 logger=None,
                 reward_min_no_blackout=-1.0,
                 reward_min_blackout=-10.0, 
                 reward_max_no_blackout=1.0,
                 reward_max_blackout=2.0,
                 reward_end_episode_bonus=1.0):
        BaseReward.__init__(self, logger=logger)
        
        self.reward_min_no_blackout : float = dt_float(reward_min_no_blackout)
        self.reward_min_blackout : float = dt_float(reward_min_blackout)
        self.reward_max_no_blackout : float = dt_float(reward_max_no_blackout)
        self.reward_max_blackout : float = dt_float(reward_max_blackout)
        self.reward_end_episode_bonus : float = dt_float(reward_end_episode_bonus)
        self.reward_no_game_over : float = dt_float(0.0)
        
        self._reward_range_blackout : float = (self.reward_max_blackout - self.reward_min_blackout)

        self.total_time_steps : Optional[int] = dt_int(0)
        self.time_window : Optional[int] = None
        
        self._ts_attack : Optional[np.ndarray] = None
        self._current_id : int = 0
        self._lines_currently_attacked : Optional[np.ndarray] = None
        self._alert_launched : Optional[np.ndarray] = None
        self._nrows_array : Optional[int] = None
        
        self._i_am_simulate : bool = False

    
    def initialize(self, env: "grid2op.Environment.BaseEnv"):
        self.total_time_steps = env.max_episode_duration()
        self.time_window = env.parameters.ALERT_TIME_WINDOW
        self._nrows_array = self.time_window + 2
        self._ts_attack = np.full((self._nrows_array, type(env).dim_alerts), False, dtype=dt_bool)
        self._alert_launched = np.full((self._nrows_array, type(env).dim_alerts), False, dtype=dt_bool)
        self._current_id = 0
        self._lines_currently_attacked = np.full(type(env).dim_alerts, False, dtype=dt_bool)
        
        self._i_am_simulate = self.is_simulated_env(env)
        return super().initialize(env)      

    def reset(self, env):
        self._ts_attack[:,:] = False
        self._alert_launched[:,:] = False
        self._current_id = 0
        self._lines_currently_attacked[:] = False
        self._i_am_simulate = self.is_simulated_env(env)
        return super().reset(env)      
    
    def _update_attack(self, env):
        if env.infos["opponent_attack_line"] is None:
            # no attack at this step
            self._lines_currently_attacked[:] = False
            self._ts_attack[self._current_id, :] = False
        else:
            # an attack at this step
            lines_attacked = env.infos["opponent_attack_line"][type(env).alertable_line_ids]
            # compute the list of lines that are "newly" attacked
            new_lines_attacked = lines_attacked & (~self._lines_currently_attacked)
            # remember the steps where these lines are attacked
            self._ts_attack[self._current_id, new_lines_attacked] = True
            # and now update the state of lines under attack
            self._lines_currently_attacked[:] = False
            self._lines_currently_attacked[lines_attacked] = True
    
    def _update_alert(self, action):
        self._alert_launched[self._current_id, :] = 1 * action.raise_alert
        
    def _update_state(self, env, action):
        self._current_id += 1
        self._current_id %= self._nrows_array
        
        # update attack
        self._update_attack(env)
        
        # update alerts
        self._update_alert(action)
        
        # update internal state of the environment
        # (this is updated in case the reward returns non 0)
        env._was_alert_used_after_attack[:] = 0
            
    def _compute_score_attack_blackout(self, env, ts_attack_in_order, indexes_to_look):
        # retrieve the lines that have been attacked in the time window
        ts_ind, line_ind = np.where(ts_attack_in_order)
        line_first_attack, first_ind_line_attacked = np.unique(line_ind, return_index=True)
        ts_first_line_attacked = ts_ind[first_ind_line_attacked]
        # now retrieve the array starting at the correct place
        ts_first_line_attacked_orig = indexes_to_look[ts_first_line_attacked]
        # and now look at the previous step if alerts were send
        # prev_ts = (ts_first_line_attacked_orig - 1) % self._nrows_array
        prev_ts = ts_first_line_attacked_orig
        # update the state of the environment
        env._was_alert_used_after_attack[line_first_attack] = self._alert_launched[prev_ts, line_first_attack] * 2 - 1
        return np.mean(self._alert_launched[prev_ts, line_first_attack]) * self._reward_range_blackout +  self.reward_min_blackout
    
    def __call__(self, action, env, has_error, is_done, is_illegal, is_ambiguous):
        # retrieve the alert made by the agent
        res = 0.
        if self._i_am_simulate:
            # does not make sense for simulate
            return res
        
        if  is_done & (not has_error): 
            # end of episode, no blackout => reward specific for this case
            return self.reward_end_episode_bonus
        
        self._update_state(env, action)
        
        if self.is_in_blackout(has_error, is_done): 
            # I am in blackout, I need to check for attack in the time window
            # if there is no attack, I do nothing
            indexes_to_look = (np.arange(-self.time_window, 1) + self._current_id) % self._nrows_array  # include current step (hence the np.arange(..., **1**))
            ts_attack_in_order = self._ts_attack[indexes_to_look, :]
            has_attack = (ts_attack_in_order).any()
            if has_attack:
                # I need to check the alarm for the attacked lines
                res = self._compute_score_attack_blackout(env, ts_attack_in_order, indexes_to_look)
        else:
            # no blackout: i check the first step in the window before me to see if there is an attack,
            index_window = (self._current_id - self.time_window) % self._nrows_array
            lines_attack = self._ts_attack[index_window, :]
            if lines_attack.any():
                # prev_ind = (index_window - 1) % self._nrows_array
                # I don't need the "-1" because the action is already BEFORE the observation in the reward.
                prev_ind = index_window
                alert_send = self._alert_launched[prev_ind, lines_attack]
                # update the state of the environment
                env._was_alert_used_after_attack[lines_attack] = 1 - alert_send * 2
                res = (self.reward_min_no_blackout - self.reward_max_no_blackout) * np.mean(alert_send) + self.reward_max_no_blackout
                self._ts_attack[index_window, :] = False  # attack has been taken into account we "cancel" it
        return res
    


# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from grid2op.VoltageControler.BaseVoltageController import BaseVoltageController


class ControlVoltageFromFile(BaseVoltageController):
    """
    This class is the most basic controler for the voltages.

    Basically, what it does is read the voltages from the
    chronics.

    If the voltages are not on the chronics (missing files), it will not change the voltage setpoint at all.
    """

    def __init__(self, gridobj, controler_backend, actionSpace_cls):
        """

        Parameters
        ----------
        gridobj: :class:`grid2op.Space.GridObjects`
            Structure of the powergrid

        controler_backend: :class:`grid2op.Backend.Backend`
            An instanciated backend to perform some computation on a powergrid, before taking some actions.

        """
        BaseVoltageController.__init__(
            self,
            gridobj=gridobj,
            controler_backend=controler_backend,
            actionSpace_cls=actionSpace_cls,
        )

    def fix_voltage(self, observation, agent_action, env_action, prod_v_chronics):
        """
        This method must be overloaded to change the behaviour of the generator setpoint for time t+1.

        This simple class will:

        - do nothing if the vector `prod_v_chronics` is None
        - set the generator setpoint to the value in prod_v_chronics

        Basically, this class is pretty fast, but does nothing interesting, beside looking at the data in files.

        More general class can use, to adapt the voltage setpoint:

        - `observation` the observation (receive by the agent) at time t
        - `agent_action` the action of the agent at time t
        - `env_action` the modification of the environment at time t, that will be observed by the agent at time
          t+1
        - `prod_v_chronics` the new setpoint of the generators present in the data (if any, this can be None)

        To help this class, a :class:`grid2op.Backend.Backend` is available and can be used to perform simulation of
        potential impact of voltages setpoints.

        Parameters
        ----------
        observation: :class:`grid2op.Observation.Observation`
            The last observation (at time t)

        agent_action: :class:`grid2op.Action.Action`
            The action that the agent took

        env_action: :class:`grid2op.Action.Action`
            The modification that the environment will take.

        prod_v_chronics: ``numpy.ndarray``
            The next voltage setpoint present in the data (if any) or ``None`` if not.

        Returns
        -------
        res: :class:`grid2op.Action.Action`
            The new setpoint, in this case depending only on the prod_v_chronics.

        """
        # TODO add a "reward" and "done" for RL voltage controler
        if prod_v_chronics is not None:
            res = self.action_space({"injection": {"prod_v": prod_v_chronics}})
        else:
            res = self.action_space()
        return res



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

"""
This module provide a simple class to represent an :class:`grid2op.BaseObservation.BaseObservation` as a plotly graph.

We are aware that the graph can be largely improved. This tool is an example on what can be done with the Grid2Op
framework.

We hope It requires an valid installation of plotly and seaborn. These dependencies can be installed with:

.. code-block:: bash

    pip3 install grid2op[plots]

To use this plotting utilities, for example in a jupyter notebook please refer to the
``getting_started/4_StudyYourAgent`` notebook for more details information. The basic usage of this function is:

.. code-block:: python

    # make the relevant import
    from grid2op.MakeEnv import make
    from grid2op.PlotPlotly import PlotObs

    # create a simple toy environment
    environment = make("case5_example")

    # set up the plot utility
    graph_layout =  [(0,0), (0,400), (200,400), (400, 400), (400, 0)]
    plot_helper = PlotObs(substation_layout=graph_layout,
                          observation_space=environment.observation_space)

    # perform a step from this environment:
    do_nothing = environment.action_space({})
    environment.step(act)

    # do the actual plot
    fig = plot_helper.get_plot_observation(environment.get_obs())
    fig.show()

"""

import numpy as np

from grid2op.Plot.BasePlot import BasePlot
from grid2op.Exceptions import PlotError

try:
    import plotly.graph_objects as go
    import seaborn as sns

    can_plot = True
except Exception as e:
    can_plot = False
    pass

# TODO add tests there


# Some utilities to plot substation, lines or get the color id for the colormap.
def draw_sub(pos, radius=50, line_color="LightSeaGreen"):
    """
    INTERNAL

    .. warning:: /!\\\\ This class is deprecated /!\\\\

    This function will draw the contour of a unique substation.

    Parameters
    ----------
    pos: ``tuple``
        It represents the position (x,y) of the center of the substation

    radius: ``float``
        Positive floating point representing the "radius" of the substation.

    Returns
    -------
    res: :class:`plotly.graph_objects.layout.Shape`
        A representation, as a plotly object of the substation

    """
    pos_x, pos_y = pos
    res = go.layout.Shape(
        type="circle",
        xref="x",
        yref="y",
        x0=pos_x - radius,
        y0=pos_y - radius,
        x1=pos_x + radius,
        y1=pos_y + radius,
        line_color=line_color,
        layer="below",
    )
    return res


def get_col(rho):
    """
    INTERNAL

    .. warning:: /!\\\\ This class is deprecated /!\\\\

    Get the index (in the color palette) of the current  capacity usage.

    Parameters
    ----------
    rho: ``float``
        The capacity usage of a given powerline.

    Returns
    -------
    res: ``int``
        The integer (between 0 and 6) of this line capacity usage in terms of color.

    """
    if rho < 0.3:
        return 0
    if rho < 0.5:
        return 1
    if rho < 0.75:
        return 2
    if rho < 0.9:
        return 3
    if rho < 0.95:
        return 5
    return 6


def draw_line(pos_sub_or, pos_sub_ex, rho, color_palette, status, line_color="gray"):
    """
    INTERNAL

    .. warning:: /!\\\\ This class is deprecated /!\\\\

    Draw a powerline with the color depending on its line capacity usage.

    Parameters
    ----------
    pos_sub_or: ``tuple``
        Position (x,y) of the origin end of the powerline

    pos_sub_ex: ``tuple``
        Position (x,y) of the extremity end of the powerline

    rho: ``float``
        Line capacity usage

    color_palette: ``object``
        The color palette to use

    status: ``bool``
        Powerline status (connected / disconnected). Disconnected powerlines are dashed.

    Returns
    -------
    res: :class:`plotly.graph_objects.layout.Shape`
        A representation, as a plotly object of the powerline

    """
    x_0, y_0 = pos_sub_or
    x_1, y_1 = pos_sub_ex

    res = go.layout.Shape(
        type="line",
        xref="x",
        yref="y",
        x0=x_0,
        y0=y_0,
        x1=x_1,
        y1=y_1,
        layer="below",
        line=dict(color=line_color, dash=None if status else "dash"),
    )
    return res


class PlotPlotly(BasePlot):
    """
    INTERNAL

    .. warning:: /!\\\\ This class is deprecated /!\\\\

        Prefer using the class `grid2op.PlotGrid.PlotPlotly`

    This class aims at simplifying the representation of an observation as a plotly object given a layout of a given
    powergrid substation.
    It "automatically" handles the positionning of the powerlines, loads and generators based on that.

    This class is just here as an inspection tool. The results can be of course improved, epsecially the label of the
    powerlines, or the ppositioning of the loads and generators.

    Attributes
    ----------
    _layout: ``dict``
        Initial layout of the powergrid.

    subs_elements: ``list``
        For each substation, it gives a representation of all the object connected to it. So, for each substation, it
        has a dictionnary with:

            - key: the name of the objects
            - value: a dictionnary representing this object containing:

                - "type" : its type, among "load", "gen" and "line"
                - "sub_pos" (``int``) and index representing which element of the substation represents this object
                - "pos" : its position as a tuple
                - "z": its position as a complex number

    cols: ``object``
        A color palette, this should not be changed for now.

    radius_sub: ``float``
        The radius of each substation. The bigger this number, the better the topology will be visible, but the more
        space taken on the overall plot

    load_prod_dist: ``float``
        The distance between a load and a generator from the center of the substation. This must be higher than
        :attr:`PlotObs.radius_sub`

    bus_radius: ``float``
        The radius of the bus. When multiple buses are present in a substation, they are materialized by a filled
        circle. This number represents the size of these circles.



    """

    def __init__(
        self,
        observation_space,
        substation_layout=None,
        radius_sub=25.0,
        load_prod_dist=70.0,
        bus_radius=4.0,
    ):
        """

        Parameters
        ----------
        substation_layout: ``list``
            List of tupe given the position of each of the substation of the powergrid.

        observation_space: :class:`grid2op.Observation.ObservationSpace`
            BaseObservation space

        """
        BasePlot.__init__(
            self,
            substation_layout=substation_layout,
            observation_space=observation_space,
            radius_sub=radius_sub,
            load_prod_dist=load_prod_dist,
            bus_radius=bus_radius,
        )
        if not can_plot:
            raise PlotError(
                'Impossible to plot as plotly cannot be imported. Please install "plotly" and '
                '"seaborn" with "pip install --update plotly seaborn"'
            )

        # define a color palette, whatever...
        sns.set()
        pal = sns.light_palette("darkred", 8)
        self.cols = pal.as_hex()[1:]

        self.col_line = "royalblue"
        self.col_sub = "red"
        self.col_load = "black"
        self.col_gen = "darkgreen"
        self.default_color = "black"
        self.type_fig_allowed = go.Figure

    def init_fig(self, fig, reward, done, timestamp):
        if fig is None:
            fig = go.Figure()
        elif not isinstance(fig, self.type_fig_allowed):
            raise PlotError(
                "PlotPlotly cannot plot on figure of type {}. The accepted type is {}. You provided an "
                'invalid argument for "fig"'.format(type(fig), self.type_fig_allowed)
            )
        return fig

    def _post_process_obs(
        self, fig, reward, done, timestamp, subs, lines, loads, gens, topos
    ):
        # update the figure with all these information
        traces = []
        subs_el = []
        lines_el = []
        loads_el = []
        gens_el = []
        topos_el = []
        for el, trace_ in subs:
            subs_el.append(el)
            traces.append(trace_)
        for el, trace_ in lines:
            lines_el.append(el)
            traces.append(trace_)
        for el, trace_ in loads:
            loads_el.append(el)
            traces.append(trace_)
        for el, trace_ in gens:
            gens_el.append(el)
            traces.append(trace_)
        for el, _ in topos:
            topos_el.append(el)
            topos_el.append(el)
            # traces.append(trace_)
        fig.update_layout(shapes=subs_el + lines_el + loads_el + gens_el + topos_el)

        for trace_ in traces:
            fig.add_trace(trace_)

        # update legend, background color, size of the plot etc.
        fig.update_xaxes(
            range=[
                np.min([el for el, _ in self._layout["substations"]])
                - 1.5 * (self.radius_sub + self.load_prod_dist),
                np.max([el for el, _ in self._layout["substations"]])
                + 1.5 * (self.radius_sub + self.load_prod_dist),
            ],
            zeroline=False,
        )
        fig.update_yaxes(
            range=[
                np.min([el for _, el in self._layout["substations"]])
                - 1.5 * (self.radius_sub + self.load_prod_dist),
                np.max([el for _, el in self._layout["substations"]])
                + 1.5 * (self.radius_sub + self.load_prod_dist),
            ]
        )
        fig.update_layout(
            margin=dict(l=20, r=20, b=100),
            height=600,
            width=800,
            plot_bgcolor="white",
            yaxis={"showgrid": False, "showline": False, "zeroline": False},
            xaxis={"showgrid": False, "showline": False, "zeroline": False},
        )
        return fig

    def _draw_subs_one_sub(self, fig, sub_id, center, this_col, txt_):
        trace = go.Scatter(
            x=[center[0]],
            y=[center[1]],
            text=[txt_],
            mode="text",
            showlegend=False,
            textfont=dict(color=this_col),
        )
        res = draw_sub(center, radius=self.radius_sub, line_color=this_col)
        return res, trace

    def _draw_powerlines_one_powerline(
        self, fig, l_id, pos_or, pos_ex, status, value, txt_, or_to_ex, this_col
    ):
        """
        INTERNAL

        .. warning:: /!\\\\ This class is deprecated /!\\\\

            Prefer using the class `grid2op.PlotGrid.PlotPlotly`

        Draw the powerline, between two substations.

        Parameters
        ----------
        observation
        fig

        Returns
        -------

        """
        tmp = draw_line(
            pos_or,
            pos_ex,
            rho=value,
            color_palette=self.cols,
            status=status,
            line_color=this_col,
        )
        trace = go.Scatter(
            x=[(pos_or[0] + pos_ex[0]) / 2],
            y=[(pos_or[1] + pos_ex[1]) / 2],
            text=[txt_],
            mode="text",
            showlegend=False,
            textfont=dict(color=this_col),
        )
        return tmp, trace

    def _draw_loads_one_load(
        self,
        fig,
        l_id,
        pos_load,
        txt_,
        pos_end_line,
        pos_load_sub,
        how_center,
        this_col,
    ):
        # add the MW load
        trace = go.Scatter(
            x=[pos_load.real],
            y=[pos_load.imag],
            text=[txt_],
            mode="text",
            showlegend=False,
            textfont=dict(color=this_col),
        )
        # add the line between the MW display and the substation
        # TODO later one, add something that looks like a load, a house for example
        res = go.layout.Shape(
            type="line",
            xref="x",
            yref="y",
            x0=pos_end_line.real,
            y0=pos_end_line.imag,
            x1=pos_load_sub[0],
            y1=pos_load_sub[1],
            layer="below",
            line=dict(color=this_col),
        )
        return res, trace

    def _draw_gens_one_gen(
        self, fig, g_id, pos_gen, txt_, pos_end_line, pos_gen_sub, how_center, this_col
    ):
        # add the MW load
        trace = go.Scatter(
            x=[pos_gen.real],
            y=[pos_gen.imag],
            text=[txt_],
            mode="text",
            showlegend=False,
            textfont=dict(color=this_col),
        )
        # add the line between the MW display and the substation
        # TODO later one, add something that looks like a generator, and could depend on the type of it!
        res = go.layout.Shape(
            type="line",
            xref="x",
            yref="y",
            x0=pos_end_line.real,
            y0=pos_end_line.imag,
            x1=pos_gen_sub[0],
            y1=pos_gen_sub[1],
            layer="below",
            line=dict(color=this_col),
        )
        return res, trace

    def _draw_topos_one_sub(self, fig, sub_id, buses_z, elements, bus_vect):
        res_sub = []
        # I plot the buses
        for bus_id, z_bus in enumerate(buses_z):
            bus_color = "#ff7f0e" if bus_id == 0 else "#1f77b4"
            res = go.layout.Shape(
                type="circle",
                xref="x",
                yref="y",
                x0=z_bus.real - self.bus_radius,
                y0=z_bus.imag - self.bus_radius,
                x1=z_bus.real + self.bus_radius,
                y1=z_bus.imag + self.bus_radius,
                fillcolor=bus_color,
                line_color=bus_color,
            )
            res_sub.append((res, None))
        # i connect every element to the proper bus with the proper color
        for el_nm, dict_el in elements.items():
            this_el_bus = bus_vect[dict_el["sub_pos"]] - 1
            if this_el_bus >= 0:
                res = go.layout.Shape(
                    type="line",
                    xref="x",
                    yref="y",
                    x0=dict_el["z"].real,
                    y0=dict_el["z"].imag,
                    x1=buses_z[this_el_bus].real,
                    y1=buses_z[this_el_bus].imag,
                    line=dict(color="#ff7f0e" if this_el_bus == 0 else "#1f77b4"),
                )
                res_sub.append((res, None))
        return res_sub

    def _get_default_cmap(self, normalized_value):
        return self.cols[get_col(normalized_value)]



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np
from grid2op.Reward.baseReward import BaseReward
from grid2op.dtypes import dt_float


class L2RPNReward(BaseReward):
    """
    This is the historical :class:`BaseReward` used for the Learning To Run a Power Network competition on WCCI 2019

    See `L2RPN <https://l2rpn.chalearn.org/>`_ for more information.

    This rewards makes the sum of the "squared margin" on each powerline.

    The margin is defined, for each powerline as:
    `margin of a powerline = (thermal limit - flow in amps) / thermal limit`
    (if flow in amps <= thermal limit) else `margin of a powerline  = 0.`

    This rewards is then: `sum (margin of this powerline) ^ 2`, for each powerline.


    Examples
    ---------
    You can use this reward in any environment with:

    .. code-block:

        import grid2op
        from grid2op.Reward import L2RPNReward

        # then you create your environment with it:
        NAME_OF_THE_ENVIRONMENT = "rte_case14_realistic"
        env = grid2op.make(NAME_OF_THE_ENVIRONMENT,reward_class=L2RPNReward)
        # and do a step with a "do nothing" action
        obs = env.reset()
        obs, reward, done, info = env.step(env.action_space())
        # the reward is computed with the L2RPNReward class

    """

    def __init__(self, logger=None):
        BaseReward.__init__(self, logger=logger)

    def initialize(self, env):
        self.reward_min = dt_float(0.0)
        self.reward_max = dt_float(env.backend.n_line)

    def __call__(self, action, env, has_error, is_done, is_illegal, is_ambiguous):
        if not is_done and not has_error:
            line_cap = self.__get_lines_capacity_usage(env)
            res = line_cap.sum()
        else:
            # no more data to consider, no powerflow has been run, reward is what it is
            res = self.reward_min
        # print(f"\t env.backend.get_line_flow(): {env.backend.get_line_flow()}")
        return res

    @staticmethod
    def __get_lines_capacity_usage(env):
        ampere_flows = np.abs(env.backend.get_line_flow(), dtype=dt_float)
        thermal_limits = np.abs(env.get_thermal_limit(), dtype=dt_float)
        thermal_limits += 1e-1  # for numerical stability
        relative_flow = np.divide(ampere_flows, thermal_limits, dtype=dt_float)

        x = np.minimum(relative_flow, dt_float(1.0))
        lines_capacity_usage_score = np.maximum(
            dt_float(1.0) - x**2, np.zeros(x.shape, dtype=dt_float)
        )
        return lines_capacity_usage_score



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
import os
import numpy as np
import json
import copy
import tempfile

from grid2op.dtypes import dt_float
from grid2op.Reward import L2RPNSandBoxScore
from grid2op.Agent import RecoPowerlineAgent
from grid2op.utils.underlying_statistics import EpisodeStatistics
from grid2op.Episode import EpisodeData

import re


class ScoreL2RPN2020(object):
    """
    This class allows to compute the same score as the one computed for the L2RPN 2020 competitions.

    It uses some "EpisodeStatistics" of the environment to compute these scores. These statistics, if not available
    are computed at the initialization.

    When using it a second time these information are reused.

    Examples
    ---------
    This class can be used as follow:

    .. code-block:: python

        import grid2op
        from grid2op.utils import ScoreL2RPN2020
        from grid2op.Agent import DoNothingAgent

        env = grid2op.make("l2rpn_case14_sandbox")
        nb_scenario = 2
        my_score = ScoreL2RPN2020(env,
                                  nb_scenario=nb_scenario,
                                  env_seeds=[0 for _ in range(nb_scenario)],
                                  agent_seeds=[0 for _ in range(nb_scenario)]
                                  )

        my_agent = DoNothingAgent(env.action_space)
        print(my_score.get(my_agent))


    Notes
    -------
    To prevent overfitting, we strongly recommend you to use the :func:`grid2op.Environment.Environment.train_val_split`
    and use this function on the built validation set only.

    Also note than computing the statistics, and evaluating an agent on a whole dataset of multiple GB can take a
    really long time and a lot of memory. This fact, again, plea in favor of using this function only on
    a validation set.

    We also strongly recommend to set the seeds of your agent (agent_seeds)
    and of the environment (env_seeds) if you want to use this feature. Reproducibility is really important if you
    want to make progress.

    .. warning::

        The triggering (or not) of the recomputation of the statistics is not perfect for now.
        We recommend you to use always
        the same seeds (`env_seeds` and `agent_seeds` key word argument of this functions)
        and the same parameters (`env.parameters`) when using a given environments.

        You might need to clean it manually if you change
        one of theses things by calling :func:`ScoreL2RPN2020.clear_all()` function .

    """

    NAME_DN = "l2rpn_dn"
    NAME_RP_NO_OVERFLOW = "l2rpn_no_overflow_reco"

    def __init__(
        self,
        env,
        env_seeds=None,
        agent_seeds=None,
        nb_scenario=16,
        min_losses_ratio=0.8,
        verbose=0,
        max_step=-1,
        nb_process_stats=1,
        scores_func=L2RPNSandBoxScore,
        score_names=None,
        add_nb_highres_sim=False,
    ):
        self.env = env
        self.nb_scenario = nb_scenario
        self.env_seeds = env_seeds
        self.agent_seeds = agent_seeds
        self.min_losses_ratio = min_losses_ratio
        self.verbose = verbose
        self.max_step = max_step

        computed_scenarios = [el[1] for el in EpisodeStatistics.list_stats(self.env)]

        self.scores_func = scores_func

        # check if i need to compute stat for do nothing
        self.stat_dn = EpisodeStatistics(self.env, self.NAME_DN)
        self._recomputed_dn = self._init_stat(
            self.stat_dn,
            self.NAME_DN,
            computed_scenarios,
            nb_process_stats=nb_process_stats,
            score_names=score_names,
        )

        # check if i need to compute that for do nothing without overflow disconnection
        param_no_overflow = copy.deepcopy(env.parameters)
        param_no_overflow.NO_OVERFLOW_DISCONNECTION = True
        # check if i need to compute that for reco powerline without overflow disconnection
        self.stat_no_overflow_rp = EpisodeStatistics(self.env, self.NAME_RP_NO_OVERFLOW)
        agent_reco = RecoPowerlineAgent(self.env.action_space)
        self._recomputed_no_ov_rp = self._init_stat(
            self.stat_no_overflow_rp,
            self.NAME_RP_NO_OVERFLOW,
            computed_scenarios,
            parameters=param_no_overflow,
            nb_process_stats=nb_process_stats,
            agent=agent_reco,
            score_names=score_names,
        )
        self.add_nb_highres_sim = add_nb_highres_sim
        self.__cleared = False 

    def _init_stat(
        self,
        stat,
        stat_name,
        computed_scenarios,
        parameters=None,
        nb_process_stats=1,
        agent=None,
        score_names=None,
    ):
        """will check if the statistics need to be computed"""
        need_recompute = True
        if score_names is None:
            score_names = [EpisodeStatistics.SCORES]
        if EpisodeStatistics.get_name_dir(stat_name) in computed_scenarios:
            # the things have been computed i check if the number of scenarios is big enough
            scores, ids_ = stat.get(score_names[0])
            metadata = stat.get_metadata()
            max_id = np.max(ids_)

            # i need to recompute if if i did not compute enough scenarios
            need_recompute = max_id < self.nb_scenario - 1

            # if max
            computed_step = int(metadata["max_step"])
            if computed_step > 0:
                # if i have computed the data with
                if self.max_step == -1:
                    # i need to compute now all the dataset, so yes i have to recompute it
                    need_recompute = True

                # i need to recompute only if i ask more steps than what was computed
                need_recompute = need_recompute or self.max_step > metadata["max_step"]

            # TODO check for the seeds here too
            # TODO and check for the class of the scores too
            # TODO check for the parameters too...

        if need_recompute:
            # i need to compute it
            if self.verbose >= 1:
                print(
                    "I need to recompute the statistics for this environment. This will take a while"
                )  # TODO logger
            stat.compute(
                nb_scenario=self.nb_scenario,
                pbar=self.verbose >= 2,
                env_seeds=self.env_seeds,
                agent_seeds=self.agent_seeds,
                scores_func=self.scores_func,
                max_step=self.max_step,
                parameters=parameters,
                nb_process=nb_process_stats,
                agent=agent,
            )
            stat.clear_episode_data()
        return need_recompute

    def _compute_episode_score(
        self,
        ep_id,  # the ID here, which is an integer and is not the ID from chronics balblabla
        meta,
        other_rewards,
        dn_metadata,
        no_ov_metadata,
        score_file_to_use=None,
    ):
        """
        Performs the rescaling of the score given the information stored in the "statistics" of this
        environment.
        """
        # load_p, ids = self.stat_no_overflow.get("load_p")
        # prod_p, _ = self.stat_no_overflow.get("prod_p")
        load_p_rp, ids_rp = self.stat_no_overflow_rp.get("load_p")
        prod_p_rp, _ = self.stat_no_overflow_rp.get("load_p")

        if score_file_to_use is None:
            score_file_to_use = EpisodeStatistics.SCORES
            key_score_file = EpisodeStatistics.KEY_SCORE
        else:
            # should match underlying_statistics.run_env `dict_kwg["other_rewards"][XXX] = ...`
            # XXX is right now f"{EpisodeStatistics.KEY_SCORE}_{nm}" [this should match the XXX]
            real_nm = EpisodeStatistics._nm_score_from_attr_name(score_file_to_use)
            key_score_file = f"{EpisodeStatistics.KEY_SCORE}_{real_nm}"

        scores_dn, ids_dn_sc = self.stat_dn.get(score_file_to_use)
        # scores_no_ov, ids_noov_sc = self.stat_no_overflow.get(score_file_to_use)
        scores_no_ov_rp, ids_noov_sc_rp = self.stat_no_overflow_rp.get(
            score_file_to_use
        )

        # reshape to have 1 dim array
        ids = ids_rp.reshape(-1)
        ids_dn_sc = ids_dn_sc.reshape(-1)
        ids_noov_sc_rp = ids_noov_sc_rp.reshape(-1)

        # there is a hugly "1" at the end of each scores due to the "game over" (or end of game), so i remove it
        scores_dn = scores_dn[ids_dn_sc == ep_id][:-1]
        scores_no_ov_rp = scores_no_ov_rp[ids_noov_sc_rp == ep_id][:-1]

        dn_this = dn_metadata[f"{ep_id}"]
        no_ov_this = no_ov_metadata[f"{ep_id}"]

        n_played = int(meta["nb_timestep_played"])
        dn_step_played = dn_this["nb_step"] - 1
        total_ts = no_ov_this["nb_step"] - 1

        ep_marginal_cost = np.max(self.env.gen_cost_per_MW).astype(dt_float)
        min_losses_ratio = self.min_losses_ratio

        # remember that first observation do not count (it's generated by the environment)
        ep_loads = load_p_rp[ids == ep_id, :].sum(axis=1)[1:]
        ep_losses = prod_p_rp[ids == ep_id, :].sum(axis=1)[1:] - ep_loads

        if self.max_step > 0:
            scores_dn = scores_dn[: self.max_step]
            # scores_no_ov = scores_no_ov[:self.max_step]
            scores_no_ov_rp = scores_no_ov_rp[: self.max_step]
            ep_loads = ep_loads[: self.max_step]
            ep_losses = ep_losses[: self.max_step]

        # do nothing operationnal cost
        ep_do_nothing_operat_cost = scores_dn.sum()
        ep_do_nothing_operat_cost += (
            ep_loads[dn_step_played:].sum() * ep_marginal_cost
        )

        # no overflow disconnection cost
        ep_do_nothing_nodisc_cost = scores_no_ov_rp.sum()

        # this agent cumulated operationnal cost
        # same as above: i remove the last element which correspond to the last state, so irrelevant
        ep_cost = np.array([el[key_score_file] for el in other_rewards]).astype(
            dt_float
        )
        if dn_metadata["max_step"] == self.max_step:
            ep_cost = ep_cost[:-1]
        ep_cost = ep_cost.sum()
        ep_cost += ep_loads[n_played:].sum() * ep_marginal_cost

        # Compute ranges
        worst_operat_cost = (
            ep_loads.sum() * ep_marginal_cost
        )  # operational cost corresponding to the min score
        zero_operat_score = ep_do_nothing_operat_cost
        nodisc_oeprat_cost = ep_do_nothing_nodisc_cost
        best_score = (
            ep_losses.sum() * min_losses_ratio
        )  # operational cost corresponding to the max score

        # Linear interp episode reward to codalab score
        if zero_operat_score != nodisc_oeprat_cost:
            # DoNothing agent doesnt complete the scenario
            reward_range = [
                best_score,
                nodisc_oeprat_cost,
                zero_operat_score,
                worst_operat_cost,
            ]
            score_range = [100.0, 80.0, 0.0, -100.0]
        else:
            # DoNothing agent can complete the scenario
            reward_range = [best_score, zero_operat_score, worst_operat_cost]
            score_range = [100.0, 0.0, -100.0]
        ep_score = np.interp(ep_cost, reward_range, score_range)
        return ep_score, n_played, total_ts

    def clear_all(self):
        """
        Has side effects

        .. warning:: /!\\\\ Be careful /!\\\\

        Clear the whole statistics directory for the 3 different computed statistics used for the score. It will
        remove the previously computed statistics.

        Once done, this cannot be undone.
        """
        # self.stat_no_overflow.clear_all()
        self.stat_no_overflow_rp.clear_all()
        self.stat_dn.clear_all()
        self.__cleared = True

    def get(self, agent, path_save=None, nb_process=1):
        """
        Get the score of the agent depending on what has been computed.

        TODO The plots will be done later.

        Parameters
        ----------
        agent: :class:`grid2op.Agent.BaseAgent`
            The agent you want to score

        path_save: ``str``
            the path were you want to store the logs of your agent.

        nb_process: ``int``
            Number of process to use for the evaluation

        Returns
        -------
        all_scores: ``list``
            List of the score of your agent per scenarios

        ts_survived: ``list``
            List of the number of step your agent successfully managed for each scenario

        total_ts: ``list``
            Total number of step for each scenario
        """
        if  self.__cleared:
            raise RuntimeError(EpisodeStatistics.ERROR_MSG_CLEANED)
        
        if path_save is not None:
            need_delete = False  # TODO this is soooo dirty
            path_save = os.path.abspath(path_save)
        else:
            need_delete = True
            dir_tmp = tempfile.TemporaryDirectory()
            path_save = dir_tmp.name
            if self.verbose >= 1:
                print("Using a temp directory to store the intermediate data.")  # TODO logger

        if self.verbose >= 1:
            print("Starts the evaluation of the agent")  # TODO logger
        nb_highres_sim = EpisodeStatistics.run_env(
            self.env,
            env_seeds=self.env_seeds,
            agent_seeds=self.agent_seeds,
            path_save=path_save,
            parameters=self.env.parameters,
            scores_func=self.scores_func,
            agent=agent,
            max_step=self.max_step,
            nb_scenario=self.nb_scenario,
            pbar=self.verbose >= 2,
            nb_process=nb_process,
            add_nb_highres_sim=self.add_nb_highres_sim,
        )
        # NB nb_highres_sim is None if self.add_nb_highres_sim is False !
        if self.verbose >= 1:
            print("Start the evaluation of the scores")  # TODO logger

        meta_data_dn = self.stat_dn.get_metadata()
        no_ov_metadata = self.stat_no_overflow_rp.get_metadata()

        all_scores = []
        ts_survived = []
        total_ts = []
        for ep_id in range(self.nb_scenario):
            this_ep_nm = meta_data_dn[f"{ep_id}"]["scenario_name"]
            with open(
                os.path.join(path_save, this_ep_nm, EpisodeData.META),
                "r",
                encoding="utf-8",
            ) as f:
                this_epi_meta = json.load(f)
            with open(
                os.path.join(path_save, this_ep_nm, EpisodeData.OTHER_REWARDS),
                "r",
                encoding="utf-8",
            ) as f:
                this_epi_scores = json.load(f)
            score_this_ep, nb_ts_survived, total_ts_tmp = self._compute_episode_score(
                ep_id,
                meta=this_epi_meta,
                other_rewards=this_epi_scores,
                dn_metadata=meta_data_dn,
                no_ov_metadata=no_ov_metadata,
            )
            all_scores.append(score_this_ep)
            ts_survived.append(nb_ts_survived)
            total_ts.append(total_ts_tmp)

        if need_delete:
            dir_tmp.cleanup()
        res = all_scores, ts_survived, total_ts
        if self.add_nb_highres_sim:
            res = all_scores, ts_survived, total_ts, nb_highres_sim
        return res


if __name__ == "__main__":
    import grid2op
    from lightsim2grid import LightSimBackend
    from grid2op.Agent import RandomAgent, DoNothingAgent

    env = grid2op.make("l2rpn_case14_sandbox", backend=LightSimBackend())
    nb_scenario = 16
    my_score = ScoreL2RPN2020(
        env,
        nb_scenario=nb_scenario,
        env_seeds=[0 for _ in range(nb_scenario)],
        agent_seeds=[0 for _ in range(nb_scenario)],
    )

    my_agent = RandomAgent(env.action_space)
    my_agent = DoNothingAgent(env.action_space)
    print(my_score.get(my_agent))



# Copyright (c) 2019-2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.


import pandapower as pp
import copy
import numpy as np

real_case = pp.from_json("grid.json")
sim_case = copy.deepcopy(real_case)

np.random.seed(42)
noise = 0.05

# change powerlines
sim_case.line["r_ohm_per_km"] *= np.random.lognormal(0., noise)
sim_case.line["x_ohm_per_km"] *= np.random.lognormal(0., noise)
# TODO do I change trafo ?

pp.runpp(sim_case)
pp.runpp(real_case)
assert sim_case.converged
assert sim_case.res_line.shape[0] == sim_case.line.shape[0]
print(f"L1 error on p: {np.mean(np.abs(sim_case.res_line['p_from_mw'] - real_case.res_line['p_from_mw'])):.2f}MW")
print(f"L1 error on q: {np.mean(np.abs(sim_case.res_line['q_from_mvar'] - real_case.res_line['q_from_mvar'])):.2f}MVAr")
pp.to_json(sim_case, "grid_forecast.json")



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np

from grid2op.Reward.l2RPNSandBoxScore import L2RPNSandBoxScore
from grid2op.dtypes import dt_float


class L2RPNWCCI2022ScoreFun(L2RPNSandBoxScore):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            It **must not** serve as a reward. This scored needs to be **MINIMIZED**,
            and a reward needs to be maximized! Also, this "reward" is not scaled or anything. Use it as your
            own risk.

    Implemented as a reward to make it easier to use in the context of the L2RPN competitions, this "reward"
    computed the "grid operation cost". It should not be used to train an agent.

    The "reward" the closest to this score is given by the :class:`RedispReward` class.
    """
    def __init__(self,
                 storage_cost=10.,  # € / MWh
                 alpha_redisp=1.0,
                 alpha_loss=1.0,
                 alpha_storage=1.0,
                 alpha_curtailment=1.0,
                 reward_max=1000.,
                 logger=None):
        super().__init__(alpha_redisp, alpha_loss, alpha_storage, alpha_curtailment, reward_max, logger)
        self.storage_cost = dt_float(storage_cost)
        
    def _get_storage_cost(self, env, p_t):
        """storage cost is a flat 10 € / MWh instead of depending on the marginal cost"""
        c_storage = np.abs(env._storage_power).sum() * self.storage_cost * env.delta_time_seconds / 3600.0
        return c_storage    



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numbers
from tarfile import ENCODING

from grid2op.Exceptions import *


def _get_default_aux(
    name,
    kwargs,
    defaultClassApp,
    _sentinel=None,
    msg_error="Error when building the default parameter",
    defaultinstance=None,
    defaultClass=None,
    build_kwargs={},
    isclass=False,
):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

    Helper to build default parameters forwarded to :class:`grid2op.Environment.Environment` for its creation.

    Exactly one of ``defaultinstance`` or ``defaultClass`` should be used, and set to not ``None``

    Parameters
    ----------
    name: ``str``
        Name of the argument to look for

    kwargs: ``dict``
        The key word arguments given to the :func:`make` function

    defaultClassApp; ``type``
        The default class to which the returned object should belong to. The final object should either be an instance
        of this ``defaultClassApp`` (if isclass is ``False``) or a subclass of this (if isclass is ``True``)

    _sentinel: ``None``
        Internal, do not use. Present to force key word arguments.

    msg_error: ``str`` or ``None``
        The message error to display if the object does not belong to ``defaultClassApp``

    defaultinstance: ``object`` or ``None``
        The default instance that will be returned. Note that if ``defaultinstance`` is not None, then
        ``defaultClass`` should be ``None`` and ``build_kwargs`` and empty dictionnary.

    defaultClass: ``type`` or ``None``
        The class used to build the default object. Note that if ``defaultClass`` is not None, then
        ``defaultinstance`` should be.

    build_kwargs:  ``dict``
        The keyword arguments used to build the final object (if ``isclass`` is ``True``). Note that:

          * if ``isclass`` is ``False``, this should be empty
          * if ``defaultinstance`` is not None, then this should be empty
          * This parameter should allow to create a valid object of type ``defaultClass``: it's key must be
            proper keys accepted by the class


    isclass: ``bool``
        Whether to build an instance of a class, or just return the class.


    Returns
    -------
    res:
        The parameters, either read from kwargs, or with its default value.

    """
    err_msg = 'Impossible to create the parameter "{}": '
    if _sentinel is not None:
        err_msg += "Impossible to get default parameters for building the environment. Please use keywords arguments."
        raise RuntimeError(err_msg)

    res = None
    # first seek for the parameter in the kwargs, and check it's valid
    if name in kwargs:
        res = kwargs[name]
        if isclass is None:
            # I don't know whether it's an object or a class
            error_msg_here = None
            res = None
            try:
                # I try to build it as an object
                res = _get_default_aux(
                    name,
                    kwargs=kwargs,
                    defaultClassApp=defaultClassApp,
                    _sentinel=_sentinel,
                    msg_error=msg_error,
                    defaultinstance=defaultinstance,
                    defaultClass=defaultClass,
                    build_kwargs=build_kwargs,
                    isclass=False,
                )
            except EnvError as exc1_:
                # I try to build it as a class
                try:
                    res = _get_default_aux(
                        name,
                        kwargs=kwargs,
                        defaultClassApp=defaultClassApp,
                        _sentinel=_sentinel,
                        msg_error=msg_error,
                        defaultinstance=defaultinstance,
                        defaultClass=defaultClass,
                        build_kwargs=build_kwargs,
                        isclass=True,
                    )
                except EnvError as exc2_:
                    # both fails !
                    error_msg_here = f"{exc1_} AND {exc2_}"

            if error_msg_here is not None:
                raise EnvError(error_msg_here)

        elif isclass is False:
            # i must create an instance of a class. I check whether it's a instance.
            if not isinstance(res, defaultClassApp):
                if issubclass(defaultClassApp, numbers.Number):
                    try:
                        # if this is base numeric type, like float or anything, i try to convert to it (i want to
                        # accept that "int" are float for example.
                        res = defaultClassApp(res)
                    except Exception as exc_:
                        # if there is any error, i raise the error message
                        raise EnvError(msg_error)
                else:
                    # if there is any error, i raise the error message
                    raise EnvError(msg_error)
        elif isclass is True:
            # so it should be a class
            if not isinstance(res, type):
                raise EnvError(
                    'Parameter "{}" should be a type and not an instance. It means that you provided an '
                    "object instead of the class to build it.".format(name)
                )
            # I must create a class, i check whether it's a subclass
            if not issubclass(res, defaultClassApp):
                raise EnvError(msg_error)
        else:
            raise EnvError(
                'Impossible to use the "_get_default_aux" function with "isclass" kwargs being different '
                "from None, True and False"
            )

    if res is None:
        # build the default parameter if not found

        if isclass is False:
            # i need building an instance
            if defaultClass is not None:
                if defaultinstance is not None:
                    err_msg += "Impossible to build an environment with both a default instance, and a default class"
                    raise EnvError(err_msg.format(name))
                try:
                    res = defaultClass(**build_kwargs)
                except Exception as e:
                    e.args = e.args + (
                        'Cannot create and instance of {} with parameters "{}"'.format(
                            defaultClass, build_kwargs
                        ),
                    )
                    raise
            elif defaultinstance is not None:
                if len(build_kwargs):
                    err_msg += "An instance is provided, yet kwargs to build it is also provided"
                    raise EnvError(err_msg.format(name))
                res = defaultinstance
            else:
                err_msg = ' None of "defaultClass" and "defaultinstance" is provided.'
                raise EnvError(err_msg.format(name))
        else:
            # I returning a class
            if len(build_kwargs):
                err_msg += (
                    "A class must be returned, yet kwargs to build it is also provided"
                )
                raise EnvError(err_msg.format(name))
            if defaultinstance is not None:
                err_msg += "A class must be returned yet a default instance is provided"
                raise EnvError(err_msg.format(name))
            res = defaultClass
    return res



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from collections.abc import Iterable
import numpy as np

from grid2op.MakeEnv import make

try:
    from lightsim2grid import LightSimBackend

    bkclass = LightSimBackend
    # raise ImportError()
except ImportError as excq_:
    from grid2op.Backend import PandaPowerBackend

    bkclass = PandaPowerBackend
    pass


class EnvCache(object):
    """
    TODO this is not implemented yet (at least the "cache" part)

    We should have a flag that when an environment is computing, it returns an "error" to indicate that on the person
    who makes the call (for now i'm pretty sure the current implementation will not work in asynch mode).
    """

    ENV_NOT_FOUND = 0
    ENV_ID_NOT_FOUND = 1
    INVALID_ACTION = 2
    INVALID_STEP = 3
    ERROR_ENV_SEED = 4
    ERROR_ENV_RESET = 5
    ERROR_ENV_SET_ID = 6
    ERROR_ENV_THERMAL_LIMIT = 7
    ERROR_CLOSE = 8
    ERROR_ENV_FAST_FORWARD = 9
    ERROR_ENV_PATH = 10

    def __init__(self, ujson_as_json):
        self.all_env = {}
        self.ujson_as_json = (
            ujson_as_json  # do i use the faster "ujson" library to parse json
        )
        self._convert_json = not self.ujson_as_json

    def insert_env(self, env_name):
        """
        TODO
        """
        try:
            env = make(env_name, backend=bkclass())  # TODO look at the RemoteEnv here
            env.deactivate_forecast()
            # call deactivate_forecast

        except Exception as exc_:
            return None, None, exc_

        if env_name not in self.all_env:
            # create an environment with that name
            self.all_env[env_name] = [env]
        else:
            self.all_env[env_name].append(env)
        id_ = len(self.all_env[env_name]) - 1
        obs = env.reset()
        return id_, obs.to_json(convert=self._convert_json), None

    def step(self, env_name, env_id, action_as_json):
        """
        TODO
        """
        res_env = (None, None, None, None)
        env, (error_id, error_msg) = self._aux_get_env(env_name, env_id)

        if error_id is not None:
            return res_env, (error_id, error_msg)
        try:
            act = env.action_space()
            act.from_json(action_as_json)
        except Exception as exc_:
            msg_ = (
                f"impossible to convert the provided action to a valid action on this environment with error:\n"
                f"{exc_}"
            )
            return res_env, (self.INVALID_ACTION, msg_)

        try:
            obs, reward, done, info = env.step(act)
        except Exception as exc_:
            msg_ = (
                f"impossible to make a step on the give environment with error\n{exc_}"
            )
            return res_env, (self.INVALID_STEP, msg_)
        return (
            obs.to_json(convert=self._convert_json),
            float(reward),
            bool(done),
            self._aux_info_to_json(info),
        ), (None, None)

    def seed(self, env_name, env_id, seed):
        """
        TODO
        """
        env, (error_id, error_msg) = self._aux_get_env(env_name, env_id)

        if error_id is not None:
            return None, (error_id, error_msg)
        try:
            seeds = env.seed(seed)
        except Exception as exc_:
            msg_ = f"Impossible to seed the environment with error:\n{exc_}"
            return None, (self.ERROR_ENV_SEED, msg_)
        return self._aux_array_to_json(seeds), (None, None)

    def reset(self, env_name, env_id):
        """
        TODO
        """
        env, (error_id, error_msg) = self._aux_get_env(env_name, env_id)
        if error_id is not None:
            return None, (error_id, error_msg)

        try:
            obs = env.reset()
        except Exception as exc_:
            msg_ = f"Impossible to reset the environment with error {exc_}"
            return None, (self.ERROR_ENV_RESET, msg_)

        return obs.to_json(convert=self._convert_json), (None, None)

    def set_id(self, env_name, env_id, chron_id):
        """
        TODO
        """
        env, (error_id, error_msg) = self._aux_get_env(env_name, env_id)
        if error_id is not None:
            return error_id, error_msg

        try:
            env.set_id(chron_id)
        except Exception as exc_:
            msg_ = f"Impossible to set the chronics id of the environment with error:\n {exc_}"
            return self.ERROR_ENV_SET_ID, msg_
        return None, None

    def set_thermal_limit(self, env_name, env_id, thermal_limit):
        """
        TODO
        """
        env, (error_id, error_msg) = self._aux_get_env(env_name, env_id)
        if error_id is not None:
            return error_id, error_msg

        try:
            env.set_thermal_limit(thermal_limit)
        except Exception as exc_:
            msg_ = f"Impossible to set the thermal limits of the environment with error:\n {exc_}"
            return self.ERROR_ENV_THERMAL_LIMIT, msg_
        return None, None

    def get_thermal_limit(self, env_name, env_id):
        """
        TODO
        """
        env, (error_id, error_msg) = self._aux_get_env(env_name, env_id)
        if error_id is not None:
            return None, (error_id, error_msg)

        try:
            res = env.get_thermal_limit()
        except Exception as exc_:
            msg_ = f"Impossible to get the thermal limits of the environment with error:\n {exc_}"
            return None, (self.ERROR_ENV_THERMAL_LIMIT, msg_)
        res = res.tolist()
        return res, (None, None)

    def close(self, env_name, env_id):
        """
        TODO
        """
        env, (error_id, error_msg) = self._aux_get_env(env_name, env_id)
        if error_id is not None:
            return error_id, error_msg

        try:
            env.close()
        except Exception as exc_:
            msg_ = f"Impossible to close the environment with error:\n {exc_}"
            return self.ERROR_CLOSE, msg_
        return None, None

    def fast_forward_chronics(self, env_name, env_id, nb_step):
        """
        TODO
        """
        env, (error_id, error_msg) = self._aux_get_env(env_name, env_id)
        if error_id is not None:
            return error_id, error_msg

        try:
            env.fast_forward_chronics(nb_step)
        except Exception as exc_:
            msg_ = f"Impossible to fast forward the environment with error:\n {exc_}"
            return self.ERROR_ENV_FAST_FORWARD, msg_
        return None, None

    def get_path_env(self, env_name, env_id):
        """
        TODO
        """
        env, (error_id, error_msg) = self._aux_get_env(env_name, env_id)
        if error_id is not None:
            return None, (error_id, error_msg)

        try:
            res = env.get_path_env()
        except Exception as exc_:
            msg_ = f"Impossible to fast forward the environment with error:\n {exc_}"
            return None, (self.ERROR_ENV_PATH, msg_)
        return res, (None, None)

    def train_val_split(self, env_name, env_id, id_chron_val):
        """
        TODO
        """
        env, (error_id, error_msg) = self._aux_get_env(env_name, env_id)
        if error_id is not None:
            return (None, None), (error_id, error_msg)

        try:
            res = env.train_val_split(
                val_scen_id=id_chron_val, add_for_train="train", add_for_val="val"
            )
        except Exception as exc_:
            msg_ = f"Impossible to split the environment with error:\n {exc_}"
            return (None, None), (self.ERROR_ENV_PATH, msg_)
        return res, (None, None)

    def _aux_array_to_json(self, array):
        if isinstance(array, Iterable):
            res = None
            if isinstance(array, np.ndarray):
                if array.shape == ():
                    res = []
            if res is None:
                res = [self._aux_array_to_json(el) for el in array]
            return res
        else:
            return float(array)

    def _aux_info_to_json(self, info):
        # TODO
        res = {}
        res["disc_lines"] = [int(el) for el in info["disc_lines"]]
        res["is_illegal"] = bool(info["is_illegal"])
        res["is_ambiguous"] = bool(info["is_ambiguous"])
        res["is_dispatching_illegal"] = bool(info["is_dispatching_illegal"])
        res["is_illegal_reco"] = bool(info["is_illegal_reco"])
        if info["opponent_attack_line"] is not None:
            res["opponent_attack_line"] = [
                bool(el) for el in info["opponent_attack_line"]
            ]
        else:
            res["opponent_attack_line"] = None
        res["exception"] = [f"{exc_}" for exc_ in info["exception"]]
        return res

    def _aux_get_env(self, env_name, env_id):
        if env_name not in self.all_env:
            return None, (
                self.ENV_NOT_FOUND,
                f'environment "{env_name}" does not exists',
            )

        li_env = self.all_env[env_name]
        env_id = int(env_id)
        nb_env = len(li_env)
        if env_id >= nb_env:
            msg_ = (
                f"you asked to run the environment {env_id}  of {env_name}. But there are only {nb_env} "
                f"such environments"
            )
            return None, (self.ENV_ID_NOT_FOUND, msg_)
        env = li_env[env_id]
        return env, (None, None)



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from grid2op.utils.l2rpn_2020_scores import ScoreL2RPN2020
from grid2op.Reward import L2RPNWCCI2022ScoreFun


class ScoreL2RPN2022(ScoreL2RPN2020):
    """This class implements the score used for the L2RPN 2022 competition, 
    taking place in the context of the WCCI 2022 competition.
    """
    def __init__(self,
                 env,
                 env_seeds=None,
                 agent_seeds=None,
                 nb_scenario=16,
                 min_losses_ratio=0.8,
                 verbose=0, max_step=-1,
                 nb_process_stats=1,
                 scores_func=L2RPNWCCI2022ScoreFun,
                 score_names=None,
                 add_nb_highres_sim=False):
        super().__init__(env,
                         env_seeds,
                         agent_seeds,
                         nb_scenario,
                         min_losses_ratio,
                         verbose,
                         max_step,
                         nb_process_stats,
                         scores_func,
                         score_names,
                         add_nb_highres_sim=add_nb_highres_sim)



# Copyright (c) 2019-2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np

from grid2op.Exceptions import (
    HandlerError
)

from grid2op.dtypes import  dt_float
from grid2op.Chronics.handlers.baseHandler import BaseHandler


class LoadQFromPHandler(BaseHandler):
    """This handler is specific for "load_q" type of data.
    
    You can use it for both "forecast" ("load_q_forecasted")
    and for environment data ("load_q").
    
    It will generate load_q based on a "q over p" ratio provided
    as input. Basically, whenever called, it will
    return (when possible): `load_q = ratio * load_p`

    .. note::
        Its current implementation heavily relies on the fact that
        when the "load_q" / "load_q_forecasted" handlers 
        are called the "load_p" / "load_p_forecasted" data are already
        computed and known.
        
    """
    def __init__(self,
                 array_name="load_q",
                 qp_ratio: float=0.7,
                 max_iter=-1):
        super().__init__(array_name, max_iter)
        if isinstance(qp_ratio, np.ndarray):
            self._qp_ratio = (1.0 * qp_ratio).astype(dt_float)
        else:
            self._qp_ratio = dt_float(qp_ratio)
        
    def done(self):
        # this is never done as long as there is a "load_p"
        return False

    def load_next(self, dict_):
        if "load_p" in dict_ and not "prod_p" in dict_ and not "prod_v" in dict_:
            if dict_["load_p"] is not None:
                return self._qp_ratio * dict_["load_p"]
        return None
    
    def initialize(self, order_backend_prods, names_chronics_to_backend):
        # nothing to do for this particular handler
        pass
    
    def check_validity(self, backend):
        if isinstance(self._qp_ratio, np.ndarray):
            if backend.n_load != self._qp_ratio.shape[0]:
                raise HandlerError(f"{self.array_name}: qp_ratio should either be a single float "
                                    "or a numpy array with as many loads as there are loads on the grid. "
                                    f"You provided {self._qp_ratio.shape[0]} ratios but there are "
                                    f"{backend.n_load} loads on the grid.")
                
    def load_next_maintenance(self):
        raise HandlerError(f"load_next_maintenance {self.array_name}: You should only "
                           "use this class for ENVIRONMENT data, and not for FORECAST data nor MAINTENANCE data. "
                          )
    
    def load_next_hazard(self):
        raise HandlerError(f"load_next_hazard {self.array_name}: You should only use "
                           "this class for ENVIRONMENT data, and not for FORECAST ")
        
    def forecast(self,
                 forecast_horizon_id,
                 inj_dict_env,
                 inj_dict_previous_forecast,
                 # eg gen_p_handler if this is set to gen_p_for_handler:
                 env_handler,  
                 # list of the 4 env handlers: (load_p_handler, load_q_handler, gen_p_handler, gen_v_handler)
                 env_handlers):
        return self.load_next(inj_dict_previous_forecast)
